# File : descriptors.py
# Author : Sébastien Deriaz
# License : GPL
#
# Descriptors are classes that describe how an adapter is connected to its device.
# Depending on the protocol, they can hold strings, integers or enums

import re
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum


class Descriptor:
    DETECTION_PATTERN = ""

    def __init__(self) -> None:
        return None

    @staticmethod
    @abstractmethod
    def from_string(string: str) -> "Descriptor":
        pass

    @abstractmethod
    def is_initialized(self) -> bool:
        pass


@dataclass
class SerialPortDescriptor(Descriptor):
    DETECTION_PATTERN = r"(COM\d+|/dev[/\w\d]+):\d+"
    port: str
    baudrate: int | None = None

    @staticmethod
    def from_string(string: str) -> "SerialPortDescriptor":
        parts = string.split(":")
        port = parts[0]
        baudrate = int(parts[1])
        return SerialPortDescriptor(port, baudrate)

    def set_default_baudrate(self, baudrate: int) -> bool:
        if self.baudrate is not None:
            self.baudrate = baudrate
            return True
        else:
            return False

    def __str__(self) -> str:
        return f"{self.port}:{self.baudrate}"

    def is_initialized(self) -> bool:
        return self.baudrate is not None


@dataclass
class IPDescriptor(Descriptor):
    class Transport(Enum):
        TCP = "TCP"
        UDP = "UDP"

        @classmethod
        def from_str(cls, value: str) -> "IPDescriptor":
            for member in cls:
                if member.value.lower() == value.lower():
                    return member  # type: ignore # TODO : Check this
            raise ValueError(f"{value} is not a valid {cls.__name__}")

    DETECTION_PATTERN = r"(\d+.\d+.\d+.\d+|[\w\.]+):\d+:(UDP|TCP)"
    address: str
    transport: Transport
    port: int | None = None
    # transport: Transport | None = None

    @staticmethod
    def from_string(string: str) -> "IPDescriptor":
        parts = string.split(":")
        address = parts[0]
        port = int(parts[1])
        transport = IPDescriptor.Transport(parts[2])
        return IPDescriptor(address, transport, port)

    def __str__(self) -> str:
        return f"{self.address}:{self.port}:{self.Transport(self.transport).value}"

    def is_initialized(self) -> bool:
        return self.port is not None and self.transport is not None


@dataclass
class VisaDescriptor(Descriptor):
    # VISA Resource Address Examples
    # GPIB (IEEE-488)
    # GPIB0::14::INSTR
    # # Serial (RS-232 or USB-Serial)
    # ASRL1::INSTR                  # Windows COM1
    # ASRL/dev/ttyUSB0::INSTR       # Linux USB serial port
    #
    # # TCPIP INSTR (LXI/VXI-11/HiSLIP-compatible instruments)
    # TCPIP0::192.168.1.100::INSTR
    # TCPIP0::my-scope.local::inst0::INSTR
    #
    # # TCPIP SOCKET (Raw TCP communication)
    # TCPIP0::192.168.1.42::5025::SOCKET
    #
    # # USB (USBTMC-compliant instruments)
    # USB0::0x0957::0x1796::MY12345678::INSTR
    #
    # # VXI (Legacy modular instruments)
    # VXI0::2::INSTR
    #
    # # PXI (Modular instrument chassis)
    # PXI0::14::INSTR
    DETECTION_PATTERN = r"([A-Z]+)(\d*|\/[^:]+)?::([^:]+)(?:::([^:]+))?(?:::([^:]+))?(?:::([^:]+))?::(INSTR|SOCKET)"

    descriptor: str

    class Interface(Enum):
        GPIB = "GPIB"
        SERIAL = "ASRL"
        TCP = "TCPIP"
        USB = "USB"
        VXI = "VXI"
        PXI = "PXI"

    @staticmethod
    def from_string(string: str) -> "VisaDescriptor":
        if re.match(VisaDescriptor.DETECTION_PATTERN, string):
            return VisaDescriptor(descriptor=string)
        else:
            raise ValueError(f"Could not parse descriptor : {string}")

    def __str__(self) -> str:
        return self.descriptor

    def is_initialized(self) -> bool:
        return True


descriptors: list[type[Descriptor]] = [
    SerialPortDescriptor,
    IPDescriptor,
    VisaDescriptor,
]


def adapter_descriptor_by_string(string_descriptor: str) -> Descriptor:
    for descriptor in descriptors:
        if re.match(descriptor.DETECTION_PATTERN, string_descriptor):
            x = descriptor.from_string(string_descriptor)
            return x
    raise ValueError(f"Could not parse descriptor string : {string_descriptor}")



# File : raw.py
# Author : Sébastien Deriaz
# License : GPL

from collections.abc import Callable
from types import EllipsisType

from syndesi.adapters.backend.adapter_backend import AdapterReadPayload, AdapterSignal
from syndesi.adapters.stop_condition import StopCondition

from ..adapters.adapter import Adapter
from ..adapters.timeout import Timeout
from .protocol import Protocol

# Raw protocols provide the user with the binary data directly,
# without converting it to string first


class Raw(Protocol):
    def __init__(
        self,
        adapter: Adapter,
        timeout: Timeout | None | EllipsisType = ...,
        event_callback: Callable[[AdapterSignal], None] | None = None,
    ) -> None:
        """
        Raw device, no presentation and application layers

        Parameters
        ----------
        adapter : IAdapter
        """
        super().__init__(
            adapter,
            timeout,
            event_callback,
        )

        # Connect the adapter if it wasn't done already
        self._adapter.connect()

    def _default_timeout(self) -> Timeout | None:
        return Timeout(response=2, action="error")

    def write(self, data: bytes) -> None:
        self._adapter.write(data)

    def query(
        self,
        data: bytes,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> bytes:
        self._adapter.flush_read()
        self.write(data)
        return self.read(timeout=timeout, stop_conditions=stop_conditions)

    def query_detailed(
        self,
        data: bytes,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> AdapterReadPayload:
        self._adapter.flush_read()
        self.write(data)
        return self.read_detailed(timeout=timeout, stop_conditions=stop_conditions)

    def read(
        self,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> bytes:
        signal = self.read_detailed(timeout=timeout, stop_conditions=stop_conditions)
        return signal.data()

    def read_detailed(
        self,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> AdapterReadPayload:
        return self._adapter.read_detailed(
            timeout=timeout, stop_conditions=stop_conditions
        )

    def _on_data_ready_event(self, data: AdapterReadPayload) -> None:
        # TODO : Call the callback here ?
        pass

    def __str__(self) -> str:
        return f"Raw({self._adapter})"



# File : ip_server.py
# Author : Sébastien Deriaz
# License : GPL
#
# IP server, WIP


# NOTE : The adapter is only meant to work with a single client, thus
# a IPServer class is made to handle the socket and generate IP adapters

# class IPServer:
#     DEFAULT_BUFFER_SIZE = 1024
#     class Protocol(Enum):
#         TCP = 'TCP'
#         UDP = 'UDP'

#     def __init__(self,
#                 port : int = None,
#                 transport : str = 'TCP',
#                 address : str = None,
#                 max_clients : int = 5,
#                 stop_condition = None,
#                 alias : str = '',
#                 buffer_size : int = DEFAULT_BUFFER_SIZE):

#         """
#         IP server adapter

#         Parameters
#         ----------
#         port : int
#             IP port. If None, the default port set with set_default_port will be used
#         transport : str
#             'TCP' or 'UDP'
#         address : str
#             Custom socket ip, None by default
#         max_clients : int
#             Maximum number of clients, 5 by default
#         stop_condition : StopCondition
#             Specify a read stop condition (None by default)
#         alias : str
#             Specify an alias for this adapter, '' by default
#         buffer_size : int
#             Socket buffer size, may be removed in the future
#         """
#         self._alias = alias
#         self._stop_condition = stop_condition
#         self._logger = logging.getLogger(LoggerAlias.ADAPTER.value)
#         self._transport = self.Protocol(transport)
#         if self._transport == self.Protocol.TCP:
#             self._logger.info("Setting up TCP IP server adapter")
#             self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#         elif self._transport == self.Protocol.UDP:
#             self._logger.info("Setting up UDP IP server adapter")
#             self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
#         else:
#             raise ValueError("Invalid protocol")

#         self._address = socket.gethostname() if address is None else address
#         self._port = port
#         self._max_clients = max_clients
#         self._opened = False


#     def set_default_port(self, port):
#         """
#         Sets IP port if no port has been set yet.

#         This way, the user can leave the port empty
#         and the driver/protocol can specify it later

#         Parameters
#         ----------
#         port : int
#         """
#         if self._port is None:
#             self._port = port


#     def open(self):
#         if self._port is None:
#             raise ValueError(f"Cannot open adapter without specifying a port")
#         self._logger.info(f"Listening to incoming connections on {self._address}:{self._port}")
#         self._socket.bind((self._address, self._port))
#         self._socket.listen(self._max_clients)
#         self._opened = True

#     def close(self):
#         if hasattr(self, '_socket'):
#             self._socket.close()
#         self._logger.info("Adapter closed !")
#         self._opened = False

#     def get_client(self,
#           stop_condition : StopConditionBackend = None,
#           timeout : Timeout = None) -> IP:
#         """
#         Wait for a client to connect to the server and return the corresponding adapter
#         """
#         if not self._opened:
#             raise RuntimeError("open() must be called before getting client")
#         client_socket, address = self._socket.accept()
#         default_timeout = Timeout(response=None,
#               continuation=IP.DEFAULT_CONTINUATION_TIMEOUT,
#               total=None)
#         return IP(_socket=client_socket, address=address,
#           stop_condition=stop_condition,
#           timeout=timeout_fuse(timeout, default_timeout))



# File : visa.py
# Author : Sébastien Deriaz
# License : GPL

"""
VISA adatper, uses a VISA backend like pyvisa-py or NI to communicate with instruments
"""

from collections.abc import Callable
from types import EllipsisType

from syndesi.adapters.backend.adapter_backend import AdapterSignal
from syndesi.adapters.stop_condition import StopCondition

from .adapter import Adapter
from .backend.descriptors import VisaDescriptor
from .timeout import Timeout


class Visa(Adapter):
    """
    VISA Adapter, allows for communication with VISA-compatible devices.
    It uses pyvisa under the hood 
    
    """
    def __init__(
        self,
        descriptor: str,
        *,
        alias: str = "",
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        timeout: None | float | Timeout | EllipsisType = ...,
        encoding: str = "utf-8",
        event_callback: Callable[[AdapterSignal], None] | None = None,
        backend_address: str | None = None,
        backend_port: int | None = None,
    ) -> None:
        super().__init__(
            descriptor=VisaDescriptor.from_string(descriptor),
            alias=alias,
            stop_conditions=stop_conditions,
            timeout=timeout,
            encoding=encoding,
            event_callback=event_callback,
            backend_address=backend_address,
            backend_port=backend_port,
        )

        self._logger.info("Setting up VISA IP adapter")

    def _default_timeout(self) -> Timeout:
        return Timeout(response=5, action="error")



# File : protocol.py
# Author : Sébastien Deriaz
# License : GPL

import logging
from abc import abstractmethod
from collections.abc import Callable
from types import EllipsisType
from typing import Any

from ..adapters.adapter import Adapter
from ..adapters.auto import auto_adapter
from ..adapters.backend.adapter_backend import AdapterReadPayload, AdapterSignal

# from syndesi.adapters.stop_condition import StopCondition
from ..adapters.timeout import Timeout
from ..tools.log_settings import LoggerAlias


class Protocol:
    def __init__(
        self,
        adapter: Adapter,
        timeout: Timeout | None | EllipsisType = ...,
        event_callback: Callable[[AdapterSignal], None] | None = None,
    ) -> None:
        # TODO : Convert the callable from AdapterSignal to ProtocolSignal or something similar
        self._adapter = auto_adapter(adapter)

        self._event_callback = event_callback
        self._adapter.set_event_callback(self.event_callback)

        if timeout is not ...:
            self._adapter.set_default_timeout(timeout)
        self._logger = logging.getLogger(LoggerAlias.PROTOCOL.value)

        if timeout is ...:
            self._adapter.set_timeout(self._default_timeout())
        else:
            self._adapter.set_timeout(timeout)

    @abstractmethod
    def _default_timeout(self) -> Timeout | None:
        pass

    def flushRead(self) -> None:
        self._adapter.flush_read()

    def event_callback(self, event: AdapterSignal) -> None:
        if self._event_callback is not None:
            self._event_callback(event)

    @abstractmethod
    def _on_data_ready_event(self, data: AdapterReadPayload) -> None:
        pass

    @abstractmethod
    def write(self, *args: Any, **kwargs: Any) -> None:
        pass

    def open(self) -> None:
        self._adapter.open()

    def try_open(self) -> bool:
        return self._adapter.try_open()

    def is_opened(self) -> bool:
        return self._adapter.is_opened()

    def close(self) -> None:
        self._adapter.close()

    @abstractmethod
    def query(
        self,
        *args: Any,
        **kwargs: Any,
        #   timeout: Union[Timeout, EllipsisType, None],
        #   stop_condition: Union[StopCondition, None, EllipsisType],
        #   full_output : bool,
    ) -> Any:
        pass

    @abstractmethod
    def read(self, *args: Any, **kwargs: Any) -> Any:
        pass



# File : adapters.py
# Author : Sébastien Deriaz
# License : GPL

"""
Adapters provide a common abstraction for the media layers (physical + data link + network)
The following classes are provided, which all are derived from the main Adapter class
  - IP
  - Serial
  - VISA

Note that technically VISA is not part of the media layer, only USB is.
This is a limitation as it is to this day not possible to communicate "raw"
with a device through USB yet

An adapter is meant to work with bytes objects but it can accept strings.
Strings will automatically be converted to bytes using utf-8 encoding
"""

import logging
import os
import queue
import subprocess
import sys
import threading
import time
import weakref
from abc import ABC, abstractmethod
from collections.abc import Callable
from enum import Enum
from multiprocessing.connection import Client, Connection
from types import EllipsisType
from typing import Any

from ..component import Component
from ..tools.errors import (
    AdapterDisconnected,
    AdapterFailedToOpen,
    AdapterTimeoutError,
    BackendCommunicationError,
)
from ..tools.types import NumberLike, is_number

from ..tools.backend_api import (
    BACKEND_PORT,
    DEFAULT_ADAPTER_OPEN_TIMEOUT,
    EXTRA_BUFFER_RESPONSE_TIME,
    Action,
    BackendResponse,
    Fragment,
    default_host,
    raise_if_error,
)
from ..tools.log_settings import LoggerAlias
from .backend.adapter_backend import (
    AdapterDisconnectedSignal,
    AdapterReadPayload,
    AdapterResponseTimeout,
    AdapterSignal,
)
from .backend.backend_tools import BACKEND_REQUEST_DEFAULT_TIMEOUT
from .backend.descriptors import Descriptor
from .stop_condition import Continuation, StopCondition, StopConditionType
from .timeout import Timeout, TimeoutAction, any_to_timeout

DEFAULT_STOP_CONDITION = Continuation(time=0.1)

DEFAULT_TIMEOUT = Timeout(response=5, action="error")

# Maximum time to let the backend start
START_TIMEOUT = 2
# Time to shutdown the backend
SHUTDOWN_DELAY = 2


class SignalQueue(queue.Queue[AdapterSignal]):
    """
    A smart queue to hold adapter signals
    """
    def __init__(self) -> None:
        self._read_payload_counter = 0
        super().__init__(0)

    def has_read_payload(self) -> bool:
        """
        Return True if the queue contains a read payload
        """
        return self._read_payload_counter > 0

    def put(self, signal: AdapterSignal) -> None:
        """
        Put a signal in the queue
        
        Parameters
        ----------
        signal : AdapterSignal
        """
        if isinstance(signal, AdapterReadPayload):
            self._read_payload_counter += 1
        return super().put(signal)

    def get(self, block: bool = True, timeout: float | None = None) -> AdapterSignal:
        """
        Get an AdapterSignal from the queue
        """
        signal = super().get(block, timeout)
        if isinstance(signal, AdapterReadPayload):
            self._read_payload_counter -= 1
        return signal

def is_backend_running(address: str, port: int) -> bool:
    """
    Return True if the backend is running
    """
    try:
        conn = Client((address, port))
    except ConnectionRefusedError:
        return False
    conn.close()
    return True

def start_backend(port: int | None = None) -> None:
    """
    Start the backend in a separate process

    A custom port can be specified

    Parameters
    ----------
    port : int
    """
    arguments = [
        sys.executable,
        "-m",
        "syndesi.adapters.backend.backend",
        "-s",
        str(SHUTDOWN_DELAY),
        "-q",
        "-p",
        str(BACKEND_PORT if port is None else port),
    ]

    stdin = subprocess.DEVNULL
    stdout = subprocess.DEVNULL
    stderr = subprocess.DEVNULL

    if os.name == "posix":
        subprocess.Popen( #pylint: disable=consider-using-with
            arguments,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            start_new_session=True,
            close_fds=True,
        )

    else:
        # Windows: detach from the parent's console so keyboard Ctrl+C won't propagate.
        create_new_process_group = subprocess.CREATE_NEW_PROCESS_GROUP  # type: ignore
        detached_process = 0x00000008  # not exposed by subprocess on all Pythons
        # Optional: CREATE_NO_WINDOW (no window even for console apps)
        create_no_window = 0x08000000

        creationflags = create_new_process_group | detached_process | create_no_window

        subprocess.Popen( #pylint: disable=consider-using-with
            arguments,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            creationflags=creationflags,
            close_fds=True,
        )


class ReadScope(Enum):
    """
    Read scope

    NEXT : Only read data after the start of the read() call
    BUFFERED : Return any data that was present before the read() call
    """
    NEXT = "next"
    BUFFERED = "buffered"


class Adapter(Component[bytes]):
    """
    Adapter class

    An adapter permits communication with a hardware device.
    The adapter is the user interface of the backend adapter
    """
    #pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        *,
        descriptor: Descriptor,
        alias: str = "",
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        timeout: Timeout | EllipsisType | NumberLike | None = ...,
        encoding: str = "utf-8",
        event_callback: Callable[[AdapterSignal], None] | None = None,
        auto_open: bool = True,
        backend_address: str | None = None,
        backend_port: int | None = None,
    ) -> None:
        """
        Adapter instance

        Parameters
        ----------
        alias : str
            The alias is used to identify the class in the logs
        timeout : float or Timeout instance
            Default timeout is
        stop_condition : StopCondition or None
            Default to None
        encoding : str
            Which encoding to use if str has to be encoded into bytes
        """
        super().__init__(LoggerAlias.ADAPTER)
        self.encoding = encoding
        self._signal_queue: SignalQueue = SignalQueue()
        self.event_callback: Callable[[AdapterSignal], None] | None = event_callback
        self.backend_connection: Connection | None = None
        self._backend_connection_lock = threading.Lock()
        self._make_backend_request_queue: queue.Queue[BackendResponse] = queue.Queue()
        self._make_backend_request_flag = threading.Event()
        self._opened = False
        self._alias = alias

        # Use custom backend or the default one
        self._backend_address = default_host if backend_address is None else backend_address
        self._backend_port = BACKEND_PORT if backend_port is None else backend_port

        # There a two possibilities here
        # A) The descriptor is fully initialized
        #    -> The adapter can be connected directly
        # B) The descriptor is not fully initialized
        #    -> Wait for the protocol to set defaults and then connect the adapter

        self.descriptor = descriptor
        self.auto_open = auto_open

        # Set the stop-condition
        self._stop_conditions: list[StopCondition]
        if stop_conditions is ...:
            self._default_stop_condition = True
            self._stop_conditions = [DEFAULT_STOP_CONDITION]
        else:
            self._default_stop_condition = False
            if isinstance(stop_conditions, StopCondition):
                self._stop_conditions = [stop_conditions]
            elif isinstance(stop_conditions, list):
                self._stop_conditions = stop_conditions
            else:
                raise ValueError("Invalid stop_conditions")

        # Set the timeout
        self.is_default_timeout = False
        self._timeout: Timeout | None
        if timeout is Ellipsis:
            # Not set
            self.is_default_timeout = True
            self._timeout = DEFAULT_TIMEOUT
        elif isinstance(timeout, Timeout):
            self._timeout = timeout
        elif is_number(timeout):
            self._timeout = Timeout(timeout, action=TimeoutAction.ERROR)
        elif timeout is None:
            self._timeout = timeout

        # Buffer for data that has been pulled from the queue but
        # not used because of termination or length stop condition
        self._previous_buffer = b""

        if self.descriptor.is_initialized():
            self.connect()

        weakref.finalize(self, self._cleanup)

        # We can auto-open only if auto_open is enabled and if
        # connection with the backend has been made (descriptor initialized)
        if self.auto_open and self.backend_connection is not None:
            self.open()

    def connect(self) -> None:
        """
        Connect to the backend
        """
        if self.backend_connection is not None:
            # No need to connect, everything has been done already
            return
        if not self.descriptor.is_initialized():
            raise RuntimeError("Descriptor wasn't initialized fully")

        if is_backend_running(self._backend_address, self._backend_port):
            self._logger.info("Backend already running")
        else:
            self._logger.info("Starting backend...")
            start_backend(self._backend_port)
            start = time.time()
            while time.time() < (start + START_TIMEOUT):
                if is_backend_running(self._backend_address, self._backend_port):
                    self._logger.info("Backend started")
                    break
                time.sleep(0.1)
            else:
                # Backend could not start
                self._logger.error("Could not start backend")

        # Create the client to communicate with the backend
        try:
            self.backend_connection = Client((default_host, BACKEND_PORT))
        except ConnectionRefusedError as err:
            raise BackendCommunicationError("Failed to connect to backend") from err
        self._read_thread = threading.Thread(
            target=self.read_thread,
            args=(self._signal_queue, self._make_backend_request_queue),
            daemon=True,
        )
        self._read_thread.start()

        # Identify ourselves
        self._make_backend_request(Action.SET_ROLE_ADAPTER)

        # Set the adapter
        self._make_backend_request(Action.SELECT_ADAPTER, str(self.descriptor))

        if self.auto_open:
            self.open()

    def _make_backend_request(
        self,
        action: Action,
        *args: Any,
        timeout: float = BACKEND_REQUEST_DEFAULT_TIMEOUT,
    ) -> BackendResponse:
        """
        Send a request to the backend and return the arguments
        """

        with self._backend_connection_lock:
            if self.backend_connection is not None:
                self.backend_connection.send((action.value, *args))

        self._make_backend_request_flag.set()
        try:
            response = self._make_backend_request_queue.get(timeout=timeout)
        except queue.Empty as err:
            raise BackendCommunicationError(
                f"Failed to receive response from backend to {action}"
            ) from err

        assert (
            isinstance(response, tuple) and len(response) > 0
        ), f"Invalid response received from backend : {response}"
        raise_if_error(response)

        return response[1:]

    def read_thread(
        self,
        signal_queue: SignalQueue,
        request_queue: queue.Queue[BackendResponse],
    ) -> None:
        """
        Main adapter thread, it constantly listens for data coming from the backend
        
        - signal -> put only the signal in the signal queue
        - otherwise -> put the whole request in the request queue
        """
        while True:
            try:
                if self.backend_connection is None:
                    raise RuntimeError("Backend connection wasn't initialized")
                response: tuple[Any, ...] = self.backend_connection.recv()
            except (EOFError, TypeError, OSError):
                signal_queue.put(AdapterDisconnectedSignal())
                request_queue.put((Action.ERROR_BACKEND_DISCONNECTED,))
                break
            else:
                if not isinstance(response, tuple):
                    raise BackendCommunicationError(
                        f"Invalid response from backend : {response}"
                    )
                action = Action(response[0])

                if action == Action.ADAPTER_SIGNAL:
                    if len(response) <= 1:
                        raise BackendCommunicationError(
                            f"Invalid event response : {response}"
                        )
                    signal: AdapterSignal = response[1]
                    if self.event_callback is not None:
                        self.event_callback(signal)
                    signal_queue.put(signal)
                else:
                    request_queue.put(response)

    @abstractmethod
    def _default_timeout(self) -> Timeout:
        pass

    def set_timeout(self, timeout: Timeout | None) -> None:
        """
        Overwrite timeout

        Parameters
        ----------
        timeout : Timeout
        """
        self._timeout = timeout

    def set_default_timeout(self, default_timeout: Timeout | None) -> None:
        """
        Set the default timeout for this adapter. If a previous
        timeout has been set, it will be fused

        Parameters
        ----------
        default_timeout : Timeout or tuple or float
        """
        if self.is_default_timeout:
            self._logger.debug(f"Setting default timeout to {default_timeout}")
            self._timeout = default_timeout

    def set_stop_conditions(
        self, stop_conditions: StopCondition | None | list[StopCondition]
    ) -> None:
        """
        Overwrite the stop-condition

        Parameters
        ----------
        stop_condition : StopCondition
        """
        if isinstance(stop_conditions, list):
            self._stop_conditions = stop_conditions
        elif isinstance(stop_conditions, StopCondition):
            self._stop_conditions = [stop_conditions]
        elif stop_conditions is None:
            self._stop_conditions = []

        self._make_backend_request(Action.SET_STOP_CONDITIONs, self._stop_conditions)

    def set_default_stop_condition(self, stop_condition: StopCondition) -> None:
        """
        Set the default stop condition for this adapter.

        Parameters
        ----------
        stop_condition : StopCondition
        """
        if self._default_stop_condition:
            self.set_stop_conditions(stop_condition)

    def flush_read(self) -> None:
        """
        Flush the input buffer
        """
        self._make_backend_request(
            Action.FLUSHREAD,
        )
        while True:
            try:
                self._signal_queue.get(block=False)
            except queue.Empty:
                break

    def previous_read_buffer_empty(self) -> bool:
        """
        Check whether the previous read buffer is empty

        Returns
        -------
        empty : bool
        """
        return self._previous_buffer == b""

    def open(self) -> None:
        """
        Start communication with the device
        """
        self._make_backend_request(
            Action.OPEN,
            self._stop_conditions,
            timeout=BACKEND_REQUEST_DEFAULT_TIMEOUT + DEFAULT_ADAPTER_OPEN_TIMEOUT,
        )
        self._logger.info("Adapter opened")
        self._opened = True

    def try_open(self) -> bool:
        try:
            self.open()
        except AdapterFailedToOpen:
            return False
        return True

    def close(self, force: bool = False) -> None:
        """
        Stop communication with the device
        """
        if force:
            self._logger.debug("Closing adapter frontend")
        else:
            self._logger.debug("Force closing adapter backend")
        self._make_backend_request(Action.CLOSE, force)

        with self._backend_connection_lock:
            if self.backend_connection is not None:
                self.backend_connection.close()

        self._opened = False

    def is_opened(self) -> bool:
        """
        Return True if the adapter is opened and False otherwise

        Returns
        -------
        opened : bool
        """
        return self._opened

    def write(self, data: bytes) -> None:
        """
        Send data to the device

        Parameters
        ----------
        data : bytes or str
        """

        if isinstance(data, str):
            data = data.encode(self.encoding)
        self._make_backend_request(Action.WRITE, data)

    def read_detailed(
        self,
        timeout: Timeout | EllipsisType | None = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        scope: str = ReadScope.BUFFERED.value,
    ) -> AdapterReadPayload:
        """
        Read data from the device

        Parameters
        ----------
        timeout : tuple, Timeout
            Temporary timeout
        stop_condition : StopCondition
            Temporary stop condition
        scope : str
            Return previous data ('buffered') or only future data ('next')
        Returns
        -------
        data : bytes
        signal : AdapterReadPayload
        """
        _scope = ReadScope(scope)
        output_signal = None
        read_timeout = None

        if timeout is ...:
            read_timeout = self._timeout
        else:
            read_timeout = any_to_timeout(timeout)

        if read_timeout is None:
            raise RuntimeError("Cannot read without setting a timeout")

        if stop_conditions is not ...:
            if isinstance(stop_conditions, StopCondition):
                stop_conditions = [stop_conditions]
            self._make_backend_request(Action.SET_STOP_CONDITIONs, stop_conditions)

        # First, we check if data is in the buffer and if the scope if set to BUFFERED
        while _scope == ReadScope.BUFFERED and self._signal_queue.has_read_payload():
            signal = self._signal_queue.get()
            if isinstance(signal, AdapterReadPayload):
                output_signal = signal
                break
            # TODO : Implement disconnect ?
        else:
            # Nothing was found, ask the backend with a START_READ request. The backend will
            # respond at most after the response_time with either data or a RESPONSE_TIMEOUT

            if not read_timeout.is_initialized():
                raise RuntimeError("Timeout needs to be initialized")

            _response = read_timeout.response()

            read_init_time = time.time()
            start_read_id = self._make_backend_request(Action.START_READ, _response)[0]

            if _response is None:
                # Wait indefinitely
                read_stop_timestamp = None
            else:
                # Wait for the response time + a bit more
                read_stop_timestamp = read_init_time + _response

            while True:
                try:
                    if read_stop_timestamp is None:
                        queue_timeout = None
                    else:
                        queue_timeout = max(
                            0,
                            read_stop_timestamp
                            - time.time()
                            + EXTRA_BUFFER_RESPONSE_TIME,
                        )

                    signal = self._signal_queue.get(timeout=queue_timeout)
                except queue.Empty as e:
                    raise BackendCommunicationError(
                        "Failed to receive response from backend"
                    ) from e
                if isinstance(signal, AdapterReadPayload):
                    output_signal = signal
                    break
                if isinstance(signal, AdapterDisconnectedSignal):
                    raise AdapterDisconnected()
                if isinstance(signal, AdapterResponseTimeout):
                    if start_read_id == signal.identifier:
                        output_signal = None
                        break
                    # Otherwise ignore it

        if output_signal is None:
            match read_timeout.action:
                case TimeoutAction.RETURN_EMPTY:
                    t = time.time()
                    return AdapterReadPayload(
                        fragments=[Fragment(b"", t)],
                        stop_timestamp=t,
                        stop_condition_type=StopConditionType.TIMEOUT,
                        previous_read_buffer_used=False,
                        response_timestamp=None,
                        response_delay=None,
                    )
                case TimeoutAction.ERROR:
                    timeout_value = read_timeout.response()
                    raise AdapterTimeoutError(
                        float("nan") if timeout_value is None else timeout_value
                    )
                case _:
                    raise NotImplementedError()

        else:
            return output_signal

    def read(
        self,
        timeout: Timeout | EllipsisType | None = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> bytes:
        signal = self.read_detailed(timeout=timeout, stop_conditions=stop_conditions)
        return signal.data()

    def _cleanup(self) -> None:
        if self._opened:
            self.close()

    def query_detailed(
        self,
        data: bytes | str,
        timeout: Timeout | EllipsisType | None = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> AdapterReadPayload:
        """
        Shortcut function that combines
        - flush_read
        - write
        - read
        """
        self.flush_read()
        self.write(data)
        return self.read_detailed(timeout=timeout, stop_conditions=stop_conditions)

    def query(
        self,
        data: bytes | str,
        timeout: Timeout | EllipsisType | None = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> bytes:
        signal = self.query_detailed(
            data=data, timeout=timeout, stop_conditions=stop_conditions
        )
        return signal.data()

    def set_event_callback(self, callback: Callable[[AdapterSignal], None]) -> None:
        self.event_callback = callback

    def __str__(self) -> str:
        return str(self.descriptor)

    def __repr__(self) -> str:
        return self.__str__()



# File : log.py
# Author : Sébastien Deriaz
# License : GPL
#
# Log utilities
# Set log level, destination, etc...

from enum import Enum


class LoggerAlias(Enum):
    ADAPTER = "syndesi.adapter"
    PROTOCOL = "syndesi.protocol"
    CLI = "syndesi.cli"
    BACKEND = "syndesi.backend"
    ADAPTER_BACKEND = "syndesi.adapter_backend"
    BACKEND_LOGGER = "syndesi.backend_logger"



# File : console.py
# Author : Sébastien Deriaz
# License : GPL

import asyncio
import os
import sys
from collections.abc import Callable
from enum import Enum
from pathlib import Path

import prompt_toolkit
from platformdirs import user_data_dir
from prompt_toolkit import PromptSession
from prompt_toolkit.application import get_app
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import FuzzyCompleter, WordCompleter
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.keys import Keys

from ..version import AUTHOR, NAME


class Shell:
    class Style(Enum):
        DEFAULT = ""
        HIGHLIGHT = "highlight"
        NOTE = "note"
        WARNING = "warning"
        _PROMPT = "prompt-marker"
        _REVERSE_SEARCH_SELECTION = "rss"
        _REVERSE_SEARCH = "rs"
        _REVERSE_SEARCH_LINE = "rs-line"
        ERROR = "error"

    HISTORY_SIZE = 10

    STYLES = {
        "highlight": "bg:#444444 #ffffff",  # background gray, foreground white
        "prompt-marker": "bold #146DA8",  # bright green marker
        "rss": "bold bg:#505050",
        "rs": "",
        Style._REVERSE_SEARCH_LINE.value: "bold bg:#505050 #ff5010",
        Style.DEFAULT.value: "",
        Style.NOTE.value: "#A0A0A0",
        Style.WARNING.value: "#ba9109",
        Style.ERROR.value: "#8f2d2d",
    }

    _toolkit_styles = prompt_toolkit.styles.Style.from_dict(STYLES)

    RS_CAPTURE_KEY = ["<any>", "up", "down", "enter"]

    DEFAULT_PROMPT = prompt_toolkit.formatted_text.FormattedText(
        [("class:" + Style._PROMPT.value, "❯ ")]
    )

    def __init__(
        self,
        on_command: Callable[[str], None],
        history_file_name: str,
        commands: list[str] | None = None,
    ):
        # Get the appropriate directory
        data_dir = Path(user_data_dir(NAME, AUTHOR))
        data_dir.mkdir(parents=True, exist_ok=True)

        # Now you can store your history file here
        history_file = data_dir / history_file_name

        self.on_command = on_command
        self._history: FileHistory = FileHistory(str(history_file))

        word_completer = WordCompleter([] if commands is None else commands)

        self.update_reverse_search_entries()
        self.session: PromptSession[str] = prompt_toolkit.PromptSession(
            history=self._history,
            auto_suggest=AutoSuggestFromHistory(),  # enables ghost suggestions
            completer=FuzzyCompleter(word_completer),
        )
        self.kb = KeyBindings()
        self.temporary_prompt: FormattedText | None = None
        self.prompt = self.DEFAULT_PROMPT
        self.reverse_search_mode = [False]
        self.reverse_search_selection = 0
        self.ask_event = asyncio.Event()
        self.ask_callback: Callable[[str], None] | None = None

        self._stop = False

        # Find a way to have access to those variables in the event nicely

        @self.kb.add("c-r")
        def _(event: KeyPressEvent) -> None:
            self.update_reverse_search_entries()
            self.update_reverse_search()
            self.reverse_search_mode[0] = True
            for c in self.RS_CAPTURE_KEY:
                self.kb.add(c)(self.reverse_search_capture)

        self.update_reverse_search_entries()

    def update_reverse_search_entries(self) -> None:
        self._reverse_search_entries = []
        for _, x in zip(
            range(self.HISTORY_SIZE), self._history.load_history_strings(), strict=False
        ):
            self._reverse_search_entries.append(x)
        # Invert because the order is newest at the bottom
        self._reverse_search_entries = self._reverse_search_entries[::-1]
        self.reverse_search_selection = len(self._reverse_search_entries) - 1

    def update_reverse_search(self) -> None:
        entries = []
        for i, command in enumerate(self._reverse_search_entries):
            selected = i == self.reverse_search_selection
            # Add the left margin
            entries.append(("class:rs_left_line", ">" if selected else " "))
            # Space
            entries.append(("class:rss" if selected else "", " "))
            # Add the command
            entries.append(("class:rss" if selected else "", f" {command}\n"))  # Style

        self.prompt = prompt_toolkit.formatted_text.FormattedText(entries)

    def exit_reverse_search(self) -> None:
        for c in self.RS_CAPTURE_KEY:
            self.kb.remove(c)
        get_app().current_buffer.document = prompt_toolkit.document.Document(
            self._reverse_search_entries[self.reverse_search_selection]
        )
        self.prompt = self.DEFAULT_PROMPT
        self.reverse_search_mode[0] = False

    def reverse_search_capture(
        self, event: prompt_toolkit.key_binding.KeyPressEvent
    ) -> None:
        key = event.key_sequence[0].key
        exit = False
        if key == Keys.Up and self.reverse_search_selection > 0:
            self.reverse_search_selection -= 1
        elif (
            key == Keys.Down
            and self.reverse_search_selection < len(self._reverse_search_entries) - 1
        ):
            self.reverse_search_selection += 1
        elif key == Keys.Enter:
            exit = True

        if exit:
            self.exit_reverse_search()
        else:
            self.update_reverse_search()
            event.app.invalidate()

    def _get_prompt(self) -> FormattedText:
        if self.temporary_prompt is not None:
            output = self.temporary_prompt
        else:
            output = self.prompt
        return output

    def stop(self) -> None:
        self._stop = True

    def run(self) -> None:
        with prompt_toolkit.patch_stdout.patch_stdout():
            while not self._stop:
                try:
                    cmd = self.session.prompt(
                        self._get_prompt,  # Allows for update of the prompt
                        key_bindings=self.kb,
                        multiline=self.reverse_search_mode[0],
                        style=self._toolkit_styles,
                    )
                    if cmd is None:
                        continue

                    command = cmd.strip()

                    if self.ask_event.is_set():
                        self.ask_event.clear()
                        if self.ask_callback is not None:
                            self.ask_callback(command)
                            self.temporary_prompt = None  # Not great here but it works
                        continue

                    if command in ["clear", "cls"]:
                        if (
                            sys.platform == "linux"
                            or sys.platform == "linux2"
                            or sys.platform == "darwin"
                        ):
                            # linux
                            os.system("clear")
                        elif sys.platform == "win32":
                            # Windows
                            os.system("cls")
                        continue
                    if command in ["exit", "quit"]:
                        break

                    self.on_command(command)

                except (KeyboardInterrupt, EOFError):
                    # print_formatted_text("\nExiting.")
                    break

    def reprompt(self, text: str | None = None) -> None:
        if self.session.app.is_running:
            if text is not None:
                self.temporary_prompt = FormattedText(
                    [("class:prompt-marker", "❯ "), ("", text)]
                )
            self.session.app.invalidate()

    def print(self, text: str, style: Style = Style.DEFAULT) -> None:
        prompt_toolkit.shortcuts.print_formatted_text(
            FormattedText([(f"class:{style.value}", text)]), style=self._toolkit_styles
        )

    def ask(self, question: str, callback: Callable[[str], None]) -> None:
        self.ask_event.set()
        self.ask_callback = callback
        self.reprompt(question)



"""
Syndesi module
"""

from .adapters.ip import IP
from .adapters.serialport import SerialPort
from .adapters.timeout import Timeout
from .adapters.visa import Visa
from .protocols.delimited import Delimited
from .protocols.modbus import Modbus
from .protocols.raw import Raw
from .protocols.scpi import SCPI
from .tools.log import log

__all__ = [
    "IP",
    "SerialPort",
    "Visa",
    "Delimited",
    "Modbus",
    "Raw",
    "SCPI",
    "log",
    "Timeout",
]



# File : errors.py
# Author : Sébastien Deriaz
# License : GPL

from pathlib import Path

from syndesi.tools.types import NumberLike

PACKAGE_PATH = Path(__file__).resolve().parent.parent


class SyndesiError(Exception):
    """Base class for all Syndesi errors"""


class BackendCommunicationError(SyndesiError):
    """Error with backend communication"""


class BackendError(SyndesiError):
    """Error inside the backend"""


class AdapterError(SyndesiError):
    """Adapter error"""

class AdapterConfigurationError(AdapterError):
    """Adapter configuration error"""

class AdapterFailedToOpen(AdapterError):
    """Adapter failed to open"""


class AdapterDisconnected(AdapterError):
    """Adapter disconnected"""


class ProtocolError(SyndesiError):
    """Protocol error"""


class AdapterTimeoutError(AdapterError):
    def __init__(self, timeout: NumberLike) -> None:
        self.timeout = timeout
        super().__init__(
            f'No response received from device within {self.timeout} seconds"'
        )


def make_error_description(e: Exception) -> str:
    tb = e.__traceback__
    if tb is None:
        error_message = ""
    else:
        while True:
            if tb.tb_next is None:
                break
            file = Path(tb.tb_next.tb_frame.f_code.co_filename).resolve()
            if not file.is_relative_to(PACKAGE_PATH):
                break
            tb = tb.tb_next

        _type = type(e)
        extra_arguments = (str(e),)
        frame = tb.tb_frame
        line_no = tb.tb_lineno
        filename = frame.f_code.co_filename
        error_message = f"{_type} : {extra_arguments} {filename}:{line_no}"

    return error_message



# File : adapter_shell.py
# Author : Sébastien Deriaz
# License : GPL


from argparse import ArgumentParser
from enum import Enum

from syndesi.adapters.adapter import Adapter
from syndesi.adapters.backend.adapter_backend import (
    AdapterDisconnectedSignal,
    AdapterReadPayload,
    AdapterSignal,
)

from ..adapters.ip import IP
from ..adapters.serialport import SerialPort
from ..adapters.timeout import Timeout
from ..adapters.visa import Visa
from ..protocols.delimited import Delimited
from ..protocols.protocol import Protocol
from ..protocols.raw import Raw
from .console import Shell

HISTORY_FILE_NAME = "syndesi"


class Format(Enum):
    TEXT = "text"
    HEX = "hex"
    BYTES = "bytes"


FORMAT_DESCRIPTION = {
    Format.TEXT: "Encoded text",
    Format.HEX: "Hex bytes either space separated or joined",
    Format.BYTES: "Use Python bytes display syntax",
}


class AdapterType(Enum):
    IP = "ip"
    SERIAL = "serial"
    VISA = "visa"


class SpecialLineEnding(Enum):
    CR = "cr"
    LF = "lf"
    CRLF = "crlf"


LINE_ENDING_CHARS = {
    SpecialLineEnding.CR: "\r",
    SpecialLineEnding.LF: "\n",
    SpecialLineEnding.CRLF: "\r\n",
}


def hex2array(raw: str) -> bytes:
    s = raw.replace(" ", "")
    if len(s) % 2 != 0:
        s = "0" + s
    try:
        array = bytes([int(s[2 * i : 2 * (i + 1)], 16) for i in range(len(s) // 2)])
    except ValueError as err:
        raise ValueError(f"Cannot parse hex string : {raw}") from err
    return array


def array2hex(array: bytes) -> str:
    return " ".join([f"{x:02X}" for x in array])


def parse_end_argument(arg: str | None) -> str | None:
    if arg is None:
        return None
    # Return a special line end char if it corresponds
    for s, t in LINE_ENDING_CHARS.items():
        if arg == s.value:
            return t
    # Otherwise parse "\\n" -> "\n"
    return arg.replace("\\n", "\n").replace("\\r", "\r")


class AdapterShell:
    DEFAULT_TERMINATION = "\n"

    def __init__(self, kind: AdapterType, input_arguments: list[str]) -> None:

        self._parser = ArgumentParser()
        self._parser.add_argument(
            "-t",
            "--timeout",
            nargs="+",
            type=float,
            required=False,
            default=[2],
            help="Adapter timeout (response)",
        )
        self._parser.add_argument(
            "-e",
            "--end",
            required=False,
            default=SpecialLineEnding.LF.value,
            help="Termination, cr, lf, crlf, none or a custom string. Only used with text format. Custom receive end can be set with --receive-end",
        )
        self._parser.add_argument(
            "--receive-end",
            required=False,
            default=None,
            help="Reception termination, same as --end but for reception only. If not set, the value of --end will be used",
        )
        self._parser.add_argument(
            "-f",
            "--format",
            default=Format.TEXT,
            help="Format, text or hex",
            choices=[x.value for x in Format],
        )
        self._parser.add_argument(
            "--backend-address", default=None, help="Address of the backend server"
        )
        self._parser.add_argument(
            "--backend-port", default=None, help="Port of the backend server"
        )

        if kind == AdapterType.IP:
            self._parser.add_argument("address", type=str)
            self._parser.add_argument("port", type=int)
            self._parser.add_argument(
                "--protocol", choices=["TCP", "UDP"], default="TCP"
            )
        elif kind == AdapterType.SERIAL:
            self._parser.add_argument("port", type=str)
            self._parser.add_argument("baudrate", type=int)
            self._parser.add_argument(
                "--rtscts", action="store_true", default=False, help="Enable RTS/CTS"
            )
        elif kind == AdapterType.VISA:
            self._parser.add_argument("descriptor", type=str)
        else:
            raise ValueError("Unsupported Kind")
        args = self._parser.parse_args(input_arguments)

        timeout = Timeout(args.timeout)

        self.adapter: Adapter
        # Create the adapter
        if kind == AdapterType.IP:
            self.adapter = IP(
                address=args.address,
                port=args.port,
                transport=args.protocol,
                timeout=timeout,
                backend_address=args.backend_address,
                backend_port=args.backend_port,
            )
        elif kind == AdapterType.SERIAL:
            self.adapter = SerialPort(
                port=args.port,
                baudrate=args.baudrate,
                timeout=timeout,
                rts_cts=args.rtscts,
                backend_address=args.backend_address,
                backend_port=args.backend_port,
            )
        elif kind == AdapterType.VISA:
            self.adapter = Visa(
                descriptor=args.descriptor,
                timeout=timeout,
                backend_address=args.backend_address,
                backend_port=args.backend_port,
            )

        self.adapter.set_default_timeout(Timeout(action="return_empty"))

        # Add the protocol
        _format = Format(args.format)
        self._protocol: Protocol
        if _format == Format.HEX:
            self._protocol = Raw(self.adapter, event_callback=self.event)
        elif _format == Format.TEXT:
            send_end = parse_end_argument(args.end)
            receive_end = parse_end_argument(args.receive_end)
            if send_end is None:
                send_end = self.DEFAULT_TERMINATION
            if receive_end is None:
                receive_end = send_end

            self._protocol = Delimited(
                self.adapter,
                termination=send_end,
                receive_termination=receive_end,
                event_callback=self.event,
            )

        # Create the shell
        self.shell = Shell(
            on_command=self.on_command,
            history_file_name=HISTORY_FILE_NAME,
            commands=[],
        )

    def run(self) -> None:
        # try:
        self.adapter.open()
        self.shell.print(
            f"Opened adapter {self.adapter.descriptor}", style=Shell.Style.NOTE
        )
        self.shell.run()

    def on_command(self, command: str) -> None:
        self._protocol.write(command)

    def event(self, signal: AdapterSignal) -> None:
        if isinstance(signal, AdapterDisconnectedSignal):

            def f(answer: str) -> None:
                if answer.lower() == "y":
                    # try:
                    self._protocol.open()
                else:
                    # Set the stop flag, exit will be effective on reprompt
                    self.shell.stop()
                self.shell.reprompt()

            self.shell.ask("Adapter disconnected, reconnect ? (y/n): ", f)
        elif isinstance(signal, AdapterReadPayload):
            data = signal.data()
            # TODO : Catch data from delimited with formatting
            self.shell.print(data.decode("ASCII"))



# File : scpi.py
# Author : Sébastien Deriaz
# License : GPL


from types import EllipsisType

from syndesi.adapters.backend.adapter_backend import AdapterReadPayload

from ..adapters.adapter import Adapter
from ..adapters.ip import IP
from ..adapters.stop_condition import StopCondition, Termination
from ..adapters.timeout import Timeout, TimeoutAction
from .protocol import Protocol


class SCPI(Protocol):
    DEFAULT_PORT = 5025

    def __init__(
        self,
        adapter: Adapter,
        send_termination: str = "\n",
        receive_termination: str | None = None,
        timeout: Timeout | None | EllipsisType = ...,
        encoding: str = "utf-8",
    ) -> None:
        """
        SDP (Syndesi Device Protocol) compatible device

        Parameters
        ----------
        adapter : Adapter
        send_termination : str
            '\n' by default
        receive_termination : str
            None by default (copy value from send_termination)
        timeout : Timeout/float/tuple
            Set device timeout
        """
        self._encoding = encoding
        # Set the default timeout

        if receive_termination is None:
            self._receive_termination = send_termination
        else:
            self._receive_termination = receive_termination
        self._send_termination = send_termination
        # Configure the adapter for stop-condition mode (timeouts will raise errors)
        if not adapter._default_stop_condition:
            raise ValueError(
                "No stop-conditions can be set for an adapter used by SCPI protocol"
            )
        adapter.set_stop_conditions(
            Termination(self._receive_termination.encode(self._encoding))
        )

        # adapter.set_timeout(self.timeout)
        if isinstance(adapter, IP):
            adapter.set_default_port(self.DEFAULT_PORT)
        # Give the adapter to the Protocol base class
        super().__init__(adapter=adapter, timeout=timeout)

        # Connect the adapter if it wasn't done already
        self._adapter.connect()

    def _default_timeout(self) -> Timeout | None:
        return Timeout(response=5, action=TimeoutAction.ERROR.value)

    def _to_bytes(self, command: str) -> bytes:
        if isinstance(command, str):
            return command.encode("ASCII")
        else:
            raise ValueError(f"Invalid command type : {type(command)}")

    def _from_bytes(self, payload: bytes) -> str:
        if isinstance(payload, bytes):
            return payload.decode("ASCII")
        else:
            raise ValueError(f"Invalid payload type : {type(payload)}")

    def _formatCommand(self, command: str) -> str:
        return command + self._send_termination

    def _unformatCommand(self, payload: str) -> str:
        return payload.replace(self._receive_termination, "")

    def _checkCommand(self, command: str) -> None:
        for c in ["\n", "\r"]:
            if c in command:
                raise ValueError(f"Invalid char {repr(c)} in command")

    def write(self, command: str) -> None:
        self._checkCommand(command)
        payload = self._to_bytes(self._formatCommand(command))
        self._adapter.write(payload)

    def write_raw(self, data: bytes, termination: bool = False) -> None:
        self._adapter.write(
            data
            + (self._send_termination.encode(self._encoding) if termination else b"")
        )

    def query(
        self,
        command: str,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> str:
        self._adapter.flush_read()
        self.write(command)

        return self.read(timeout=timeout, stop_conditions=stop_conditions)

    def read_raw(
        self,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> bytes:
        signal = self.read_detailed(
            timeout=timeout,
            stop_conditions=stop_conditions,
        )
        return signal.data()

    def read_detailed(
        self,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> AdapterReadPayload:

        signal = self._adapter.read_detailed(
            timeout=timeout, stop_conditions=stop_conditions
        )
        return signal

    def read(
        self,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> str:
        signal = self.read_detailed(timeout=timeout, stop_conditions=stop_conditions)
        raw_data = signal.data()
        return self._unformatCommand(self._from_bytes(raw_data))



import socket
from multiprocessing.connection import Connection

BACKEND_REQUEST_DEFAULT_TIMEOUT = 0.5


def get_conn_addresses(conn: Connection) -> tuple[tuple[str, int], tuple[str, int]]:
    try:
        fd = conn.fileno()
    except OSError:
        return (("closed", 0), ("closed", 0))
    else:
        sock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
        # try:
        # TODO : Implement exception
        # address, port = sock.getpeername()  # (IP, port) tuple
        peer_address = sock.getpeername()
        sock_address = sock.getsockname()
        return sock_address, peer_address
        # except Exception:
        #     return (("error", 0), ("closed", 0))


class ConnectionDescriptor:
    def __init__(self, conn: Connection) -> None:
        """
        Description of a multiprocessing Connection
        """
        local, remote = get_conn_addresses(conn)
        self._remote_address = remote[0]
        self._remote_port = int(remote[1])
        self._local_address = local[0]
        self._local_port = int(local[1])

    def remote(self) -> str:
        return f"{self._remote_address}:{self._remote_port}"

    def local(self) -> str:
        return f"{self._local_address}:{self._local_port}"

    def remote_address(self) -> str:
        return self._remote_address

    def remote_port(self) -> int:
        return self._remote_port

    def local_address(self) -> str:
        return self._local_address

    def local_port(self) -> int:
        return self._local_port

    def __str__(self) -> str:
        return f"{self.local()}->{self.remote()}"


class NamedConnection(ConnectionDescriptor):
    def __init__(self, conn: Connection) -> None:
        super().__init__(conn)
        self.conn = conn

    def __str__(self) -> str:
        return f"Connection {self.remote()}"

    def __repr__(self) -> str:
        return self.__str__()



# File : syndesi_backend.py
# Author : Sébastien Deriaz
# License : GPL

import argparse

from ..cli.backend_console import BackendConsole
from ..cli.backend_status import BackendStatus
from ..cli.backend_wrapper import BackendWrapper


def main() -> None:
    argument_parser = argparse.ArgumentParser(prog="syndesi-backend")

    argument_parser.add_argument(
        "--status", action="store_true", help="Show backend status"
    )

    argument_parser.add_argument(
        "--console", action="store_true", help="Run backend console"
    )

    args, remaining_args = argument_parser.parse_known_args()

    if args.status:
        status = BackendStatus(remaining_args)
        status.run()
    elif args.console:
        console = BackendConsole(remaining_args)
        console.run()
    else:
        backend = BackendWrapper(remaining_args)
        backend.run()


if __name__ == "__main__":
    main()



# File : backendclient.py
# Author : Sébastien Deriaz
# License : GPL
#
# The backend client manages the link between clients (frontend) and the adapter (backend)
# It is instanctiated by the backend and has a thread to manage incoming data from clients
# as well as read incoming data from the adapter backend

import logging
import threading
import time
from enum import Enum
from multiprocessing.connection import Pipe, wait
from typing import Any

from syndesi.adapters.backend.stop_condition_backend import (
    stop_condition_to_backend,
)
from syndesi.tools.errors import AdapterError, make_error_description
from syndesi.tools.types import NumberLike

from ...tools.backend_api import Action, frontend_send
from ...tools.log_settings import LoggerAlias
from .adapter_backend import (
    AdapterBackend,
    Selectable,
    nmin,
)
from .backend_tools import NamedConnection
from .descriptors import (
    Descriptor,
    IPDescriptor,
    SerialPortDescriptor,
    VisaDescriptor,
    adapter_descriptor_by_string,
)
from .ip_backend import IPBackend
from .serialport_backend import SerialPortBackend

# from .stop_condition_backend import stop_condition_from_list
from .visa_backend import VisaBackend


class TimeoutEvent(Enum):
    MONITORING = 0
    ADAPTER = 1
    # CONNECTIONS = 2


def get_adapter(descriptor: Descriptor) -> AdapterBackend:
    # The adapter doesn't exist, create it
    if isinstance(
        descriptor, SerialPortDescriptor
    ):  # Add mandatory timeout and stop_condition here ?
        return SerialPortBackend(descriptor=descriptor)
    elif isinstance(descriptor, IPDescriptor):
        return IPBackend(descriptor=descriptor)
    elif isinstance(descriptor, VisaDescriptor):
        return VisaBackend(descriptor=descriptor)
    else:
        raise ValueError(f"Unsupported descriptor : {descriptor}")


class AdapterSession(threading.Thread):
    MONITORING_DELAY = 0.5
    daemon = True
    _shutdown_counter_top: int | None
    _shutdown_counter: int | None

    def __init__(self, adapter_descriptor: str, shutdown_delay: NumberLike | None):
        super().__init__(daemon=True)
        self._logger = logging.getLogger(LoggerAlias.ADAPTER_BACKEND.value)
        self._logger.setLevel("DEBUG")
        self._role = None
        self._next_monitoring_timestamp = time.time() + self.MONITORING_DELAY

        # self._stop_flag = False
        self._connections_lock = threading.Lock()
        # self._connection_condition = threading.Condition(self._connections_lock)

        descriptor = adapter_descriptor_by_string(adapter_descriptor)

        self._adapter: AdapterBackend = get_adapter(descriptor)

        self.connections: list[NamedConnection] = []
        # self.connection_names : Dict[NamedConnection] = {}

        # self._new_connection_r, self._new_connection_w = os.pipe()
        # os.pipe does not work on Windows
        self._new_connection_r, self._new_connection_w = Pipe()

        self._shutdown_delay = shutdown_delay
        if self._shutdown_delay is not None:
            self._shutdown_counter_top = int(
                round(self._shutdown_delay / self.MONITORING_DELAY)
            )
            self._shutdown_counter = self._shutdown_counter_top
        else:
            self._shutdown_counter_top = None
            self._shutdown_counter = None

        self._read_init_id = 0

    def add_connection(self, conn: NamedConnection) -> None:
        with self._connections_lock:
            self.connections.append(conn)
            self._new_connection_w.send(b"\x00")

    def _remove_connection(self, conn: NamedConnection) -> None:
        with self._connections_lock:
            if conn in self.connections:
                conn.conn.close()
                self.connections.remove(conn)

    def send(self, conn: NamedConnection, action: Action, *args: Any) -> None:
        if not frontend_send(conn.conn, action, *args):
            self._logger.warning(f"Failed to send to {conn.remote()}")
            self._remove_connection(conn)

    def send_to_all(self, action: Action, *args: Any) -> None:
        for conn in self.connections:
            frontend_send(conn.conn, action, *args)

    def enumerate_connections(self) -> list[str]:
        return [x.remote_address() for x in self.connections]

    def is_adapter_opened(self) -> bool:
        return self._adapter.is_opened()

    def run(self) -> None:
        while True:
            try:
                stop = self.loop()
                if stop:
                    break
            except Exception as e:
                error_message = make_error_description(e)

                self._logger.critical(
                    f"Error in {self._adapter.descriptor} session loop : {error_message}"
                )
                try:
                    error_message = make_error_description(e)

                    for conn in self.connections:
                        frontend_send(conn.conn, Action.ERROR_GENERIC, error_message)
                except Exception:
                    break
        self._logger.info(f"Exit {self._adapter.descriptor} session loop")

    def loop(self) -> bool:
        # This is the main loop of the session
        # It listens for the following events :
        # - New client
        #   -> asynchronous from backend
        # - Event on a current client connection
        #   -> listen to conn
        # - Adapter event
        #   -> listen to socket/fd
        # The wait has a timeout set by the adapter, it corresponds to the current continuation/total timeout

        # Create a list of what is awaited
        wait_list: list[Selectable] = [conn.conn for conn in self.connections]
        adapter_fd = self._adapter.selectable()
        if adapter_fd is not None and adapter_fd.fileno() >= 0:
            wait_list.append(adapter_fd)

        wait_list.append(self._new_connection_r)

        timeout_timestamp = None
        event = None

        adapter_timestamp = self._adapter.get_next_timeout()
        if adapter_timestamp is not None:
            timeout_timestamp = nmin(timeout_timestamp, adapter_timestamp)
            event = TimeoutEvent.ADAPTER

        if (
            timeout_timestamp is None
            or self._next_monitoring_timestamp < timeout_timestamp
        ):
            timeout_timestamp = self._next_monitoring_timestamp
            event = TimeoutEvent.MONITORING

        if timeout_timestamp is None:
            timeout = None
        else:
            timeout = timeout_timestamp - time.time()
        ready = wait(wait_list, timeout=timeout)  # type: ignore
        t = time.time()
        if len(ready) == 0:
            # Timeout event
            if event == TimeoutEvent.MONITORING:
                self._next_monitoring_timestamp = t + self.MONITORING_DELAY
                stop = self._monitor()
                if stop:
                    return True
            elif event == TimeoutEvent.ADAPTER:
                signal = self._adapter.on_timeout_event()
                if signal is not None:
                    # The signal can be none if it has been disabled in the meantime
                    self._logger.debug(f"Adapter signal (timeout) : {signal}")
                    self.send_to_all(Action.ADAPTER_SIGNAL, signal)

        # Main adapter loop
        if self._new_connection_r in ready:
            # New connection event
            self._new_connection_r.recv()
        # Adapter event
        if self._adapter.selectable() in ready:
            for signal in self._adapter.on_socket_ready():
                self._logger.debug(f"Adapter signal (selectable) : {signal}")
                self.send_to_all(Action.ADAPTER_SIGNAL, signal)

        for conn in self.connections:
            if conn.conn in ready:
                # Manage a command received from the user
                self.manage_conn(conn)
        return False

    def _monitor(self) -> bool:
        stop = False
        if self._shutdown_counter is not None:
            with self._connections_lock:
                if len(self.connections) == 0:
                    if self._shutdown_counter == 0:
                        # Shutdown
                        self._logger.info(
                            f"No clients on adapter {self._adapter.descriptor} for {self._shutdown_delay}s, closing"
                        )
                        self._adapter.close()
                        stop = True
                    else:
                        self._shutdown_counter -= 1
                else:
                    self._shutdown_counter = self._shutdown_counter_top

        return stop

    def manage_conn(self, conn: NamedConnection) -> None:
        extra_arguments: tuple[Any, ...]
        remove_after_response = False
        if not conn.conn.poll():
            # No data, connection is closed
            self._logger.warning(f"Client {conn.remote()} closed unexpectedly")
            self._remove_connection(conn)
            return
        try:
            request = conn.conn.recv()
        except (EOFError, ConnectionResetError) as e:
            # Probably a ping or an error
            self._logger.warning(
                f"Failed to read from client {conn.remote()} ({e}), closing connection "
            )
            self._remove_connection(conn)
        else:
            if not (isinstance(request, tuple) and len(request) >= 1):
                response_action = Action.ERROR_INVALID_REQUEST
                extra_arguments = ("",)
            else:
                action: Action
                action = Action(request[0])
                response_action = Action.ERROR_GENERIC
                extra_arguments = ("Unknown error in session",)
                try:
                    match action:
                        case Action.OPEN:
                            self._adapter.set_stop_conditions(
                                [stop_condition_to_backend(sc) for sc in request[1]]
                            )
                            try:
                                self._adapter.open()
                            except AdapterError as e:
                                # Success !
                                response_action = Action.ERROR_FAILED_TO_OPEN
                                extra_arguments = (str(e),)
                            else:
                                response_action = Action.OPEN
                                extra_arguments = ("",)
                        case Action.WRITE:
                            data = request[1]
                            if self._adapter.is_opened():
                                if self._adapter.write(data):
                                    # Success
                                    response_action, extra_arguments = Action.WRITE, ()
                                else:
                                    response_action, extra_arguments = (
                                        Action.ERROR_ADAPTER_DISCONNECTED,
                                        ("",),
                                    )
                                    # TODO : Maybe close here ? not sure
                            else:
                                response_action, extra_arguments = (
                                    Action.ERROR_ADAPTER_NOT_OPENED,
                                    ("Open adapter before writing",),
                                )
                                self._logger.error("Could not write, adapter is closed")
                        case Action.PING:
                            response_action, extra_arguments = Action.PING, ()
                        case Action.SET_STOP_CONDITIONs:
                            self._adapter.set_stop_conditions(
                                [stop_condition_to_backend(sc) for sc in request[1]]
                            )
                            response_action, extra_arguments = (
                                Action.SET_STOP_CONDITIONs,
                                (),
                            )
                        case Action.FLUSHREAD:
                            self._adapter.flush_read()
                            response_action, extra_arguments = Action.FLUSHREAD, ()
                        case Action.START_READ:
                            response_time = float(request[1])
                            self._adapter.start_read(response_time, self._read_init_id)
                            response_action, extra_arguments = Action.START_READ, (
                                self._read_init_id,
                            )
                            self._read_init_id += 1

                        # case Action.GET_BACKEND_TIME:
                        #     response_action = Action.GET_BACKEND_TIME
                        #     extra_arguments = (request_timestamp, )
                        case Action.CLOSE:
                            force = request[1]
                            # Close this connection
                            remove_after_response = True
                            response_action, extra_arguments = Action.CLOSE, ()
                            if force:
                                self._adapter.close()
                        case _:
                            response_action, extra_arguments = (
                                Action.ERROR_UNKNOWN_ACTION,
                                (f"{action}",),
                            )
                except Exception as e:
                    error_message = make_error_description(e)

                    response_action, extra_arguments = (
                        Action.ERROR_GENERIC,
                        (error_message,),
                    )

            frontend_send(conn.conn, response_action, *extra_arguments)
            if remove_after_response:
                self._logger.info(f"Closing client {conn.remote()} connection")
                self._remove_connection(conn)



# File : exceptions
# Author : Sébastien Deriaz
# License : GPL


class SyndesiException(Exception):
    def __init__(self, *args: object) -> None:
        """
        Syndesi base exception class
        """
        super().__init__(*args)


class SyndesiTimeoutException(SyndesiException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



# File : delimited.py
# Author : Sébastien Deriaz
# License : GPL

from collections.abc import Callable
from types import EllipsisType

from ..adapters.adapter import Adapter
from ..adapters.backend.adapter_backend import AdapterReadPayload, AdapterSignal
from ..adapters.stop_condition import StopCondition, Termination
from ..adapters.timeout import Timeout
from .protocol import Protocol


class Delimited(Protocol):
    def __init__(
        self,
        adapter: Adapter,
        termination: str = "\n",
        format_response: bool = True,
        encoding: str = "utf-8",
        timeout: Timeout | None | EllipsisType = ...,
        event_callback: Callable[[AdapterSignal], None] | None = None,
        receive_termination: str | None = None,
    ) -> None:
        """
        Protocol with delimiter, like LF, CR, etc... LF is used by default

        No presentation or application layers

        Parameters
        ----------
        adapter : Adapter
        termination : bytes
            Command termination, '\\n' by default
        format_response : bool
            Apply formatting to the response (i.e removing the termination), True by default
        encoding : str or None
            If None, delimited will not encode/decode
        timeout : Timeout
            None by default (default timeout)
        receive_termination : bytes
            Termination when receiving only, optional
            if not set, the value of termination is used
        """
        if not isinstance(termination, str) or isinstance(termination, bytes):
            raise ValueError(
                f"end argument must be of type str or bytes, not {type(termination)}"
            )
        if receive_termination is None:
            self._receive_termination = termination
        else:
            self._receive_termination = receive_termination
        self._termination = termination
        self._encoding = encoding
        self._response_formatting = format_response

        adapter.set_stop_conditions(
            stop_conditions=Termination(sequence=self._receive_termination)
        )
        super().__init__(adapter, timeout=timeout, event_callback=event_callback)

        # Connect the adapter if it wasn't done already
        self._adapter.connect()

        # TODO : Disable encoding/decoding when encoding==None

    def __str__(self) -> str:
        if self._receive_termination == self._termination:
            return f"Delimited({self._adapter},{repr(self._termination)})"
        else:
            return f"Delimited({self._adapter},{repr(self._termination)}/{repr(self._receive_termination)})"

    def _default_timeout(self) -> Timeout | None:
        return Timeout(response=2, action="error")

    def __repr__(self) -> str:
        return self.__str__()

    def _to_bytes(self, command: str | bytes) -> bytes:
        if isinstance(command, str):
            return command.encode("ASCII")
        elif isinstance(command, bytes):
            return command
        else:
            raise ValueError(f"Invalid command type : {type(command)}")

    def _from_bytes(self, payload: bytes) -> str:
        assert isinstance(payload, bytes)
        return payload.decode("ASCII")  # TODO : encoding ?

    def _format_command(self, command: str) -> str:
        return command + self._termination

    def _format_response(self, response: str) -> str:
        if response.endswith(self._receive_termination):
            response = response[: -len(self._receive_termination)]
        return response

    def _on_data_ready_event(self, data: AdapterReadPayload) -> None:
        # TODO : Call the callback here ?
        # output = self._format_read(data.data(), decode=True)
        # return output
        pass

    def write(self, command: str) -> None:
        command = self._format_command(command)
        self._adapter.write(self._to_bytes(command))

    def query(self, data: str, timeout: Timeout | None | EllipsisType = ...) -> str:
        """
        Writes then reads from the device and return the result

        Parameters
        ----------
        data : str
            Data to send to the device
        timeout : Timeout
            Custom timeout for this query (optional)
        decode : bool
            Decode incoming data, True by default
        full_output : bool
            return metrics on read operation (False by default)
        """
        self._adapter.flush_read()
        self.write(data)
        return self.read(timeout=timeout)

    def read_raw(
        self,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> bytes:
        """
        Reads command and formats it as a str

        Parameters
        ----------
        timeout : Timeout
        decode : bool
            Decode incoming data, True by default
        full_output : bool
            If True, Return data and read information in a additionnal BackendReadOutput class
            If False, Return data only
        """

        # Send up to the termination
        signal = self._adapter.read_detailed(
            timeout=timeout, stop_conditions=stop_conditions
        )
        return signal.data()

    def read_detailed(
        self,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> AdapterReadPayload:
        signal = self._adapter.read_detailed(
            timeout=timeout, stop_conditions=stop_conditions
        )
        return signal

    def read(
        self,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
    ) -> str:
        signal = self.read_detailed(timeout=timeout, stop_conditions=stop_conditions)
        return self._decode(signal.data())

    def _decode(self, data: bytes) -> str:
        try:
            data_string = data.decode(self._encoding)
        except UnicodeDecodeError as err:
            raise ValueError(
                f"Failed to decode {data!r} to {self._encoding} ({err})"
            ) from err
        else:
            if not self._response_formatting:
                # Add the termination back in since it was removed by the adapter
                data_string += self._receive_termination

        return data_string



# File : stop_condition.py
# Author : Sébastien Deriaz
# License : GPL
#
# A stop-condition describes when a communication with a device should
# be stopped based on the data received (length, contents, termination, etc...)
# A stop-condition can also format the data if necessary (remove termination for example)

import time

from syndesi.adapters.stop_condition import (
    Continuation,
    Length,
    StopCondition,
    StopConditionType,
    Termination,
    Total,
)

from ...tools.backend_api import Fragment


def termination_in_data(termination: bytes, data: bytes) -> tuple[int | None, int]:
    """
    Return the position (if it exists) and length of the termination (or part of it) inside data
    """
    p = None
    L = len(termination)
    # First check if the full termination is somewhere. If that's the case, data will be split
    try:
        p = data.index(termination)
        # If found, return that
    except ValueError:
        # If not, we'll try to find if part of the sequence is at the end, in that case
        # we'll return the length of the sequence that was found
        L -= 1
        while L > 0:
            if data[-L:] == termination[:L]:
                p = len(data) - L  # - 1
                break
            L -= 1

    return p, L


class StopConditionBackend:
    def __init__(self) -> None:
        pass

    def initiate_read(self) -> None:
        raise NotImplementedError()

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        raise NotImplementedError()

    def type(self) -> StopConditionType:
        raise NotImplementedError()

    def flush_read(self) -> None:
        raise NotImplementedError()


class TerminationBackend(StopConditionBackend):
    def __init__(self, sequence: bytes | str) -> None:
        super().__init__()
        if isinstance(sequence, str):
            self._sequence = sequence.encode('utf-8')
        else:
            self._sequence = sequence
        self._sequence_found_length = 0

    def initiate_read(self) -> None:
        # Termination
        self._sequence_found_length = 0

    def flush_read(self) -> None:
        self._sequence_found_length = 0

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        if raw_fragment.data is None:
            raise RuntimeError("Trying to evaluate an invalid fragment")

        position, length = termination_in_data(
            self._sequence[self._sequence_found_length :], raw_fragment.data
        )
        stop = False
        deferred = Fragment(b"", None)

        if position is None:
            # Nothing was found, keep everything
            kept = raw_fragment
        else:
            self._sequence_found_length += length

            if self._sequence_found_length == len(self._sequence):
                # The sequence was found entirely
                deferred = raw_fragment[position + length :]
                self._sequence_found_length = 0
                stop = True
            elif position + length == len(raw_fragment.data):
                # Part of the sequence was found at the end
                # Return what's before the sequence
                deferred = Fragment(b"", None)

            kept = raw_fragment[:position]

        return stop, kept, deferred, None

    def type(self) -> StopConditionType:
        return StopConditionType.TERMINATION


class LengthBackend(StopConditionBackend):
    def __init__(self, N: int) -> None:
        super().__init__()
        self._N = N
        self._counter = 0

    def initiate_read(self) -> None:
        # Length
        self._counter = 0

    def flush_read(self) -> None:
        self._counter = 0

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        remaining_bytes = self._N - self._counter
        kept_fragment = raw_fragment[:remaining_bytes]
        deferred_fragment = raw_fragment[remaining_bytes:]
        self._counter += len(kept_fragment.data)
        remaining_bytes = self._N - self._counter
        # TODO : remaining_bytes <= 0 ? Alongside above TODO maybe
        return remaining_bytes == 0, kept_fragment, deferred_fragment, None

    def type(self) -> StopConditionType:
        return StopConditionType.LENGTH


class ContinuationBackend(StopConditionBackend):
    def __init__(self, time: float) -> None:
        super().__init__()
        self._continuation = time
        self._last_fragment: float | None = None

    def initiate_read(self) -> None:
        self._last_fragment = time.time()

    def flush_read(self) -> None:
        self._last_fragment = None

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        deferred = Fragment(b"", None)
        kept = raw_fragment

        if raw_fragment.timestamp is None:
            raise RuntimeError("Cannot evaluate fragment with no timestamp")
        # last_fragment can be none if no data was ever received
        if self._last_fragment is not None:
            continuation_timestamp = self._last_fragment + self._continuation
            stop = continuation_timestamp <= raw_fragment.timestamp
            next_event_timeout = continuation_timestamp
        else:
            stop = False
            next_event_timeout = None

        return stop, kept, deferred, next_event_timeout

    def type(self) -> StopConditionType:
        return StopConditionType.TIMEOUT


class TotalBackend(StopConditionBackend):
    def __init__(self, time: float) -> None:
        super().__init__()
        self._total = time
        self._start_time: float | None = None

    def initiate_read(self) -> None:
        self._start_time = time.time()

    def flush_read(self) -> None:
        self._start_time = None

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        kept = raw_fragment
        deferred = Fragment(b"", None)

        if raw_fragment.timestamp is None:
            raise RuntimeError("Cannot evaluate fragment with no timestamp")

        if self._start_time is None:
            raise RuntimeError("Invalid start time")
        total_timestamp = self._start_time + self._total
        stop = total_timestamp <= raw_fragment.timestamp

        return stop, kept, deferred, total_timestamp

    def type(self) -> StopConditionType:
        return StopConditionType.TIMEOUT


def stop_condition_to_backend(stop_condition: StopCondition) -> StopConditionBackend:
    if isinstance(stop_condition, Termination):
        return TerminationBackend(stop_condition.sequence)
    elif isinstance(stop_condition, Length):
        return LengthBackend(stop_condition.n)
    elif isinstance(stop_condition, Continuation):
        return ContinuationBackend(stop_condition.time)
    elif isinstance(stop_condition, Total):
        return TotalBackend(stop_condition.time)
    else:
        raise RuntimeError(f"Invalid stop condition : {stop_condition}")



# File : backend_logger.py
# Author : Sébastien Deriaz
# License : GPL
#
# This class starts a thread to grab log records from the backend and emit them here
# in their respective loggers

import logging
import threading
from multiprocessing.connection import Client
from time import sleep

from syndesi.adapters.backend.backend_tools import NamedConnection
from syndesi.tools.errors import BackendCommunicationError
from syndesi.tools.log_settings import LoggerAlias

from .backend_api import BACKEND_PORT, Action, backend_request, default_host


class BackendLogger(threading.Thread):
    def __init__(self) -> None:
        self._conn_description_lock = threading.Lock()
        self.conn_description = ""
        self._logger = logging.getLogger(LoggerAlias.BACKEND_LOGGER.value)
        self._logger.setLevel(logging.DEBUG)
        super().__init__(daemon=True)

    def run(self) -> None:
        conn = None
        loggers: dict[str, logging.Logger] = {}
        while True:
            if conn is None:
                try:
                    conn = NamedConnection(Client((default_host, BACKEND_PORT)))
                except ConnectionRefusedError:
                    conn = None
                    sleep(0.1)
                    continue
                else:
                    with self._conn_description_lock:
                        self.conn_description = conn.local()
                    self._logger.info("Backend connected")

                try:
                    backend_request(conn.conn, Action.SET_ROLE_LOGGER)
                except BackendCommunicationError:
                    conn.conn.close()
                    conn = None
                    sleep(0.1)
                    continue

            else:
                try:
                    record: logging.LogRecord = conn.conn.recv()
                except (EOFError, OSError, Exception):
                    self._logger.info("Backend disconnected")
                    sleep(0.1)
                    conn.conn.close()
                    conn = None
                else:
                    logger_name = record.name
                    if logger_name not in loggers:
                        loggers[logger_name] = logging.getLogger(logger_name)
                    loggers[logger_name].handle(record)



# File : visa.py
# Author : Sébastien Deriaz
# License : GPL
#
# The VISA backend communicates using pyvisa

from __future__ import annotations

import queue
import socket
import threading
import time
from types import ModuleType
from typing import TYPE_CHECKING

from ...tools.backend_api import AdapterBackendStatus, Fragment
from .adapter_backend import (
    AdapterBackend,
    AdapterDisconnectedSignal,
    AdapterSignal,
    HasFileno,
)
from .descriptors import VisaDescriptor

# --- Typing-only imports so mypy knows pyvisa symbols without requiring it at runtime
if TYPE_CHECKING:
    import pyvisa  # type: ignore
    from pyvisa.resources import Resource  # type: ignore

# --- Runtime optional import
try:
    import pyvisa as _pyvisa_runtime
except Exception:
    _pyvisa_runtime = None

pyvisa: ModuleType | None = _pyvisa_runtime  # type: ignore


class VisaBackend(AdapterBackend):
    def __init__(self, descriptor: VisaDescriptor):
        """
        USB VISA stack adapter
        """
        super().__init__(descriptor=descriptor)
        self.descriptor: VisaDescriptor

        if pyvisa is None:
            raise ImportError(
                "Missing optional dependency 'pyvisa'. Install with:\n"
                "  python -m pip install pyvisa"
            )

        # Safe: guarded above, so mypy knows pyvisa is a module here
        self._rm = pyvisa.ResourceManager()
        self._inst: Resource | None = None  # annotation only; no runtime import needed

        # We need a socket pair because VISA doesn't expose a selectable fileno/socket
        # So we create a thread to read data and push that to the socket
        self._notify_recv, self._notify_send = socket.socketpair()
        self._notify_recv.setblocking(False)
        self._notify_send.setblocking(False)

        self._stop_lock = threading.Lock()
        self.stop = False

        self._fragment_lock = threading.Lock()
        self._fragment = Fragment(b"", None)
        self._event_queue: queue.Queue[AdapterSignal] = queue.Queue()

    @classmethod
    def list_devices(cls: type[VisaBackend]) -> list[str]:
        """
        Returns a list of available VISA devices
        """
        if pyvisa is None:
            raise ImportError(
                "Missing optional dependency 'pyvisa'. Install with:\n"
                "  python -m pip install pyvisa"
            )

        rm = pyvisa.ResourceManager()
        available_resources: list[str] = []
        for device in rm.list_resources():
            try:
                d = rm.open_resource(device)
                d.close()
                available_resources.append(device)
            except pyvisa.VisaIOError:
                # Device cannot be opened; skip it
                pass

        return available_resources

    def flush_read(self) -> bool:
        super().flush_read()
        while not self._event_queue.empty():
            self._event_queue.get()
        return True

    def open(self) -> bool:
        output = False
        if self._inst is None:
            # NOTE: self._rm is always defined in __init__ when pyvisa is present
            self._inst = self._rm.open_resource(self.descriptor.descriptor)

        if self._status == AdapterBackendStatus.DISCONNECTED:
            # These attributes exist on pyvisa resources
            self._inst.write_termination = ""
            self._inst.read_termination = None

            self._inst_lock = threading.Lock()
            self._status = AdapterBackendStatus.CONNECTED

        if self._thread is None:
            self._thread = threading.Thread(
                target=self._internal_thread,
                args=(self._inst, self._event_queue),
                daemon=True,
            )
            self._thread.start()
            output = True

        return output

    def close(self) -> bool:
        super().close()
        with self._inst_lock:
            if self._inst is not None:
                self._inst.close()
        self._status = AdapterBackendStatus.DISCONNECTED
        with self._stop_lock:
            self.stop = True
        return True

    def write(self, data: bytes) -> bool:
        super().write(data)
        with self._inst_lock:
            if self._inst is not None:
                self._inst.write_raw(data)
        return True

    def _socket_read(self) -> Fragment:
        self._notify_recv.recv(1)
        if not self._event_queue.empty():
            event = self._event_queue.get()
            if isinstance(event, AdapterDisconnectedSignal):
                return Fragment(b"", None)

        with self._fragment_lock:
            output = self._fragment
            self._fragment = Fragment(b"", None)
            return output

    def _internal_thread(
        self,
        inst: Resource,
        event_queue: queue.Queue[AdapterSignal],
    ) -> None:
        assert pyvisa is not None  # for type-narrowing inside this method
        timeout = 2000
        while True:
            payload = b""
            with self._fragment_lock:
                self._fragment = Fragment(b"", None)
            try:
                inst.timeout = timeout
            except pyvisa.InvalidSession:
                pass
            try:
                while True:
                    # Read up to an error
                    payload += inst.read_bytes(1)
                    inst.timeout = 0
            except pyvisa.VisaIOError:
                # Timeout
                if payload:
                    with self._fragment_lock:
                        if self._fragment.timestamp is None:
                            self._fragment.timestamp = time.time()
                            self._fragment.data += payload
                    # Tell the session that there's data (write to a virtual socket)
                    self._notify_send.send(b"1")
            except (TypeError, pyvisa.InvalidSession, BrokenPipeError):
                event_queue.put(AdapterDisconnectedSignal())
                self._notify_send.send(b"1")
            with self._stop_lock:
                if self.stop:
                    break

    def selectable(self) -> HasFileno | None:
        return self._notify_recv

    def is_opened(self) -> bool:
        if self._inst is None:
            return False
        else:
            return self._status == AdapterBackendStatus.CONNECTED



# File : timeout.py
# Author : Sébastien Deriaz
# License : GPL
#
# The timeout backend manages timing when reading from a device, it both
# records how much time each action took and stops communication if it
# exceeds set criteria

# from enum import Enum
# from typing import Union, Tuple
# from time import time
# import json

# class TimeoutAction(Enum):
#     DISCARD = "discard"  # If a timeout is reached, data is discarded
#     RETURN = "return"  #  If a timeout is reached, data is returned (timeout acts as a stop condition)
#     STORE = "store"  # If a timeout is reached, data is stored and returned on the next read() call
#     ERROR = "error"  # If a timeout is reached, raise an error


# class TimeoutType(Enum):
#     RESPONSE = "response"

# class JsonKey(Enum):
#     RESPONSE = "r"
#     ACTION = "a"


# class TimeoutBackend:
#     class _State(Enum):
#         WAIT_FOR_RESPONSE = 0
#         CONTINUATION = 1

#     def __init__(
#         self, response, continuation, total, on_response, on_continuation, on_total
#     ) -> None:
#         """
#         A class to manage timeouts

#         Timeouts are split in three categories :
#         - response timeout : the "standard" timeout, (i.e the time it takes for
#             a device to start transmitting data)
#         - continuation timeout : the time between reception of
#             data units (bytes or blocks of data)
#         - total timeout : maximum time from start to end of transmission.
#             This timeout can stop a communication mid-way. It is used to
#             prevent a host from getting stuck reading a constantly streaming device

#         Each timeout is specified in seconds

#         Actions
#         - discard : Discard all of data obtained during the read() call if the specified timeout if reached and return empty data
#         - return : Return all of the data read up to this point when the specified timeout is reached
#         - store : Store all of the data read up to this point in a buffer. Data will be available at the next read() call
#         - error : Produce an error

#         Parameters
#         ----------
#         response : float
#         continuation : float
#         total : float
#         on_response : str
#             Action on response timeout (see Actions)
#         on_continuation : str
#             Action on continuation timeout (see Actions)
#         on_total : str
#             Action on total timeout (see Actions)
#         """
#         super().__init__()
#         # It is possible to pass a tuple to set response/continuation/total, parse this first if it is the case
#         if isinstance(response, (tuple, list)):
#             if len(response) >= 3:
#                 total = response[2]
#             if len(response) >= 2:
#                 continuation = response[1]
#             response = response[0]

#         # Timeout values (response, continuation and total)
#         self._response = response
#         self._continuation = continuation
#         self._total = total
#         self._on_response = (
#             TimeoutAction(on_response) if on_response is not None else None
#         )
#         self._on_continuation = (
#             TimeoutAction(on_continuation) if on_continuation is not None else None
#         )
#         self._on_total = TimeoutAction(on_total) if on_total is not None else None

#         # State machine flags
#         self._state = self._State.WAIT_FOR_RESPONSE
#         self._queue_timeout_type = TimeoutType.RESPONSE
#         self._last_data_action_origin = TimeoutType.RESPONSE

#     def initiate_read(self, deferred_buffer: bool = False) -> Union[float, None]:
#         """
#         Initiate a read sequence.

#         The maximum time that should be spent in the next byte read
#         is returned

#         Returns
#         -------
#         stop : bool
#             Timeout is reached
#         keep : bool
#             True if data read up to this point should be kept
#             False if data should be discarded
#         timeout : float or None
#             None is there's no timeout
#         """
#         self._start_time = time()
#         if deferred_buffer:
#             self._state = self._State.CONTINUATION
#             self._data_action = self._on_continuation
#             self._queue_timeout_type = TimeoutType.CONTINUATION
#         else:
#             self._state = self._State.WAIT_FOR_RESPONSE
#             self._data_action = self._on_response
#             self._queue_timeout_type = TimeoutType.RESPONSE
#         self._last_timestamp = None

#         self.response_time = None
#         self.continuation_times = []
#         self.total_time = None

#         self._output_timeout = None

#         return self._response

#     def evaluate(self, timestamp: float) -> Tuple[bool, Union[float, None]]:
#         stop = False

#         # self._check_uninitialization()

#         self._data_action = None
#         self._stop_source_value = (
#             "###"  # When a timeout occurs, store the value that exceeded its value here
#         )
#         self._stop_source_limit = "###"  # And store the limit value here

#         # First check if the timestamp is None, that would mean the timeout was reached in the queue
#         if timestamp is None:
#             # Set the data action according to the timeout that was given for the queue before
#             match self._queue_timeout_type:
#                 case TimeoutType.RESPONSE:
#                     self._data_action = self._on_response
#                     self._stop_source_limit = self._response
#                 case TimeoutType.CONTINUATION:
#                     self._data_action = self._on_continuation
#                     self._stop_source_limit = self._continuation
#                 case TimeoutType.TOTAL:
#                     self._data_action = self._on_total
#                     self._stop_source_limit = self._total
#             self._stop_source_value = None  # We do not have the exceed time, but None will be printed as '---'
#             self._last_data_action_origin = self._queue_timeout_type
#             stop = True

#         else:
#             # Check total
#             if self._total is not None:
#                 self.total_time = timestamp - self._start_time
#                 if self.total_time >= self._total:
#                     stop = True
#                     self._data_action = self._on_total
#                     self._last_data_action_origin = TimeoutType.TOTAL
#                     self._stop_source_value = self.total_time
#                     self._stop_source_limit = self._total
#             # Check continuation
#             elif (
#                 self._continuation is not None
#                 and self._state == self._State.CONTINUATION
#                 and self._last_timestamp is not None
#             ):
#                 continuation_time = timestamp - self._last_timestamp
#                 self.continuation_times.append(continuation_time)
#                 if continuation_time >= self._continuation:
#                     stop = True
#                     self._data_action = self._on_continuation
#                     self._last_data_action_origin = TimeoutType.CONTINUATION
#                     self._stop_source_value = continuation_time
#                     self._stop_source_limit = self._continuation
#             # Check response time
#             elif (
#                 self._response is not None
#                 and self._state == self._State.WAIT_FOR_RESPONSE
#             ):
#                 self.response_time = timestamp - self._start_time
#                 if self.response_time >= self._response:
#                     stop = True
#                     self._data_action = self._on_response
#                     self._last_data_action_origin = TimeoutType.RESPONSE
#                     self._stop_source_value = self.response_time
#                     self._stop_source_limit = self._response

#         self._output_timeout = None
#         # If we continue
#         if not stop:
#             # Update the state
#             if self._state == self._State.WAIT_FOR_RESPONSE:
#                 self._state = self._State.CONTINUATION
#             self._last_timestamp = timestamp
#             # No timeouts were reached, return the next one
#             # Return the timeout (state is always CONTINUATION at this stage)
#             # Take the smallest between continuation and total
#             if self._total is not None and self._continuation is not None:
#                 c = self._continuation
#                 t = self._start_time + self._total
#                 if c < t:
#                     self._output_timeout = c
#                     self._queue_timeout_type = TimeoutType.CONTINUATION
#                 else:
#                     self._output_timeout = t
#                     self._queue_timeout_type = TimeoutType.TOTAL
#             elif self._total is not None:
#                 self._output_timeout = time() - (self._start_time + self._total)
#                 self._queue_timeout_type = TimeoutType.TOTAL
#             elif self._continuation is not None:
#                 self._output_timeout = self._continuation
#                 self._queue_timeout_type = TimeoutType.CONTINUATION

#         return stop, self._output_timeout

#     def dataAction(self):
#         """
#         Return the data action (discard, return, store or error)
#         and the timeout origin (response, continuation or total)

#         Returns
#         -------
#         data_action : Timeout.DataAction
#         origin : Timeout.TimeoutType

#         """
#         return self._data_action, self._last_data_action_origin

#     def __str__(self) -> str:
#         def _format(value, action):
#             if value is None:
#                 return "None"
#             elif value is Ellipsis:
#                 return "not set"
#             else:
#                 return f'{value*1e3:.3f}ms/{action.value if isinstance(action, Enum) else "not set"}'

#         response = "r:" + _format(self._response, self._on_response)
#         continuation = "c:" + _format(self._continuation, self._on_continuation)
#         total = "t:" + _format(self._total, self._on_total)
#         return f"Timeout({response},{continuation},{total})"

#     def __repr__(self) -> str:
#         return self.__str__()



# File : ip.py
# Author : Sébastien Deriaz
# License : GPL

"""
IP adapter, communicates with TCP or UDP
"""
from collections.abc import Callable
from types import EllipsisType

from syndesi.adapters.backend.adapter_backend import AdapterSignal
from syndesi.tools.types import NumberLike

from .adapter import Adapter
from .backend.descriptors import IPDescriptor
from .stop_condition import StopCondition
from .timeout import Timeout

# TODO : Server ? create an adapter from a socket ?

# TODO : Manage opening and closing, modes ? open at instance or at write/read ?
# close after read ? error if already opened before / strict mode ?


class IP(Adapter):
    """
    IP Adapter, allows for communication with IP devices through UDP or TCP
    
    Parameters
    ----------
    address : str
        IP description
    port : int
        IP port
    transport : str
        'TCP' or 'UDP'
    timeout : Timeout | float
        Specify communication timeout
    stop_condition : StopCondition
        Specify a read stop condition (None by default)
    auto_open : bool
        Automatically open the adapter
    socket : socket.socket
        Specify a custom socket, this is reserved for server application
    
    
    """
    DEFAULT_PROTOCOL = IPDescriptor.Transport.TCP

    def __init__(
        self,
        address: str,
        port: int | None = None,
        transport: str = DEFAULT_PROTOCOL.value,
        *,
        timeout: Timeout | NumberLike | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        alias: str = "",
        encoding: str = "utf-8",
        event_callback: Callable[[AdapterSignal], None] | None = None,
        auto_open: bool = True,
        backend_address: str | None = None,
        backend_port: int | None = None,
    ):
        """
        Instanciate new IP adapter
        """
        super().__init__(
            descriptor=IPDescriptor(
                address=address,
                port=port,
                transport=IPDescriptor.Transport(transport.upper()),
            ),
            alias=alias,
            timeout=timeout,
            stop_conditions=stop_conditions,
            encoding=encoding,
            event_callback=event_callback,
            auto_open=auto_open,
            backend_address=backend_address,
            backend_port=backend_port,
        )
        self.descriptor: IPDescriptor

        # if self.descriptor.transport is not None:
        self._logger.info(f"Setting up {self.descriptor.transport.value} IP adapter")

        self.set_default_timeout(self._default_timeout())

    def _default_timeout(self) -> Timeout:
        return Timeout(response=5, action="error")

    def set_default_port(self, port: int) -> None:
        """
        Sets IP port if no port has been set yet.

        This way, the user can leave the port empty
        and the driver/protocol can specify it later

        Parameters
        ----------
        port : int
        """
        if self.descriptor.port is None:
            self.descriptor.port = port

    def set_default_transport(self, transport: str | IPDescriptor.Transport) -> None:
        """
        Sets the default IP transport protocol

        Parameters
        ----------
        transport : str | IPDescriptor.Transport
        """
        if self.descriptor.transport is None:
            self.descriptor.transport = IPDescriptor.Transport(transport)


# File : serialport.py
# Author : Sébastien Deriaz
# License : GPL
#
# The SerialPort backend communicates with serial devices using the serial package

import time

import serial
from serial.serialutil import PortNotOpenError, SerialException

from syndesi.tools.backend_api import AdapterBackendStatus, Fragment
from syndesi.tools.errors import AdapterConfigurationError, AdapterFailedToOpen

from .adapter_backend import AdapterBackend, HasFileno
from .descriptors import SerialPortDescriptor


class SerialPortBackend(AdapterBackend):
    def __init__(self, descriptor: SerialPortDescriptor):
        """
        Serial communication adapter

        Parameters
        ----------
        port : str
            Serial port (COMx or ttyACMx)
        """
        super().__init__(descriptor=descriptor)
        self.descriptor: SerialPortDescriptor
        self._logger.info(f"Setting up SerialPort adapter {self.descriptor}")
        self._port: serial.Serial | None = None
        self._rts_cts = False

    def set_baudrate(self, baudrate: int) -> None:
        """
        Set baudrate

        Parameters
        ----------
        baudrate : int
        """
        self.descriptor.baudrate = baudrate

        if self.is_opened():
            self.close()
            self.open()

    # TODO : Check if this is still necessary, it's probably more of a frontend thing
    def set_default_baudrate(self, baudrate: int) -> None:
        """
        Sets the default baudrate

        Parameters
        ----------
        baudrate : int
        """
        self.descriptor.set_default_baudrate(baudrate)

    def flush_read(self) -> bool:
        super().flush_read()
        if self._port is None:
            return False
        else:
            self._port.flush()
            return True

    def set_rtscts(self, enabled: bool) -> None:
        self._rts_cts = enabled
        if self.is_opened():
            self.close()
            self.open()

    def open(self) -> None:
        if self.descriptor.baudrate is None:
            raise AdapterConfigurationError("Baudrate must be set, please use set_baudrate")

        if self._port is None:
            try:
                self._port = serial.Serial(
                    port=self.descriptor.port,
                    baudrate=self.descriptor.baudrate,
                    rtscts=self._rts_cts,
                )
            except SerialException as e:
                if 'No such file' in str(e):
                    raise AdapterFailedToOpen(f"No such file or directory '{self.descriptor.port}'") from e
                else:
                    raise AdapterFailedToOpen('Unknown error') from e
            
        elif not self._port.isOpen():  # type: ignore
            self._port.open()

        if self._port.isOpen():  # type: ignore
            self._logger.info(f"Adapter {self.descriptor} opened")
            self._status = AdapterBackendStatus.CONNECTED
        else:
            self._logger.error(f"Failed to open adapter {self.descriptor}")
            self._status = AdapterBackendStatus.DISCONNECTED
            raise AdapterFailedToOpen('Unknown error')

    def close(self) -> bool:
        super().close()
        if hasattr(self, "_port") and self._port is not None:
            # Close and the read thread will die by itself
            self._port.close()
            self._logger.info("Adapter closed")
            return True
        else:
            return False

    def write(self, data: bytes) -> bool:
        super().write(data)
        if self._port is None:
            self._logger.error(f"Cannot write to closed adapter {self.descriptor}")
            return False
        else:
            if self._rts_cts:  # Experimental
                self._port.setRTS(True)  # type: ignore
            # TODO : Implement auto open
            # if self._status == AdapterBackendStatus.DISCONNECTED:
            #     self.open()
            # write_start = time.time()
            try:
                self._port.write(data)
            except (OSError, PortNotOpenError):
                return False
            #            write_duration = time.time() - write_start
            # self._logger.debug(f"Write [{write_duration * 1e3:.3f}ms]: {repr(data)}")

            return True

    def _socket_read(self) -> Fragment:
        t = time.time()

        if self._port is None:
            self._logger.debug('Port is None -> b""')
            return Fragment(b"", t)
        else:
            try:
                data = self._port.read_all()
            except (OSError, PortNotOpenError):
                self._logger.debug('Port error -> b""')
                return Fragment(b"", t)
            else:
                # This is a test, it seems b"" happens sometimes and disconnects the adapter
                if data is not None and data != b"":
                    return Fragment(data, t)
                else:
                    return Fragment(b"", t)
                    # self._logger.debug('Data is none -> b""')
                # else:
                #     self._logger.debug(f'{data=}')

    # def _start_thread(self):
    #     """
    #     Start the read thread
    #     """
    #     #super()._start_thread()
    #     if self._thread is None or not self._thread.is_alive():
    #         self._thread = Thread(
    #             target=self._read_thread,
    #             daemon=True,
    #             args=(self._port, self._read_queue, self._thread_commands_read),
    #         )
    #         self._thread.start()

    # def _read_thread(
    #     self,
    #     port: serial.Serial,
    #     read_queue: TimedQueue,
    #     thread_commands: socket.socket,
    # ):
    #     # On linux, it is possivle to use the select.select for both serial port and stop socketpair.
    #     # On Windows, this is not possible. so the port timeout is used instead.
    #     if sys.platform == "win32":
    #         port.timeout = 0.1
    #     while True:
    #         # Check how many bytes are available
    #         if sys.platform == "win32":
    #             ready, _, _ = select.select([thread_commands], [], [], 0)
    #             if thread_commands in ready:
    #                 # Stop the read thread
    #                 break
    #             else:
    #                 # Read data from the serialport with a timeout, if the timeout occurs, read again.
    #                 # This is to avoid having a crazy fast loop
    #                 fragment = port.read()
    #                 if len(fragment) > 0:
    #                     read_queue.put(fragment)
    #         else:
    #             ready, _, _ = select.select([self._port.fd, thread_commands], [], [])
    #             if thread_commands in ready:
    #                 command = self.ThreadCommands(thread_commands.recv(1))
    #                 if command == self.ThreadCommands.STOP:
    #                     # Stop the read thread
    #                     break
    #             elif self._port.fd in ready:
    #                 try:
    #                     in_waiting = self._port.in_waiting
    #                 except OSError:
    #                     # Input/output error, the port was disconnected
    #                     read_queue.put(AdapterDisconnected)
    #                 else:
    #                     fragment = port.read(in_waiting)
    #                     if fragment:
    #                         read_queue.put(fragment)

    # def read(
    #     self, timeout=..., stop_condition=..., full_output: bool = False
    # ) -> bytes:
    #     """
    #     Read data from the device

    #     Parameters
    #     ----------
    #     timeout : Timeout or None
    #         Set a custom timeout, if None (default), the adapter timeout is used
    #     stop_condition : StopCondition or None
    #         Set a custom stop condition, if None (Default), the adapater stop condition is used
    #     full_output : bool
    #         Return a dictionary containing information about the read operation like
    #             'read_duration' : float
    #             'origin' : 'timeout' or 'stop_condition'
    #             'timeout' : Timeout.TimeoutType
    #             'stop_condition' : Length or Termination (StopCondition class)
    #             'previous_read_buffer_used' : bool
    #             'n_fragments' : int
    #     """
    #     output = super().read(timeout, stop_condition, full_output)
    #     if self._rts_cts:  # Experimental
    #         self._port.setRTS(False)
    #     return output

    def is_opened(self) -> bool:
        if self._port is not None:
            if self._port.isOpen():  # type: ignore
                return True

        return False

    def selectable(self) -> HasFileno | None:
        if self._port is not None and self.is_opened():
            return self._port
        else:
            return None



# backend_status.py
# 13.07.2025
# Sébastien Deriaz

import argparse
import logging
import threading
import time
from collections.abc import Callable
from multiprocessing.connection import Client, Pipe, wait
from time import sleep
from typing import cast

from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from syndesi.adapters.backend.backend_tools import NamedConnection
from syndesi.tools.log_settings import LoggerAlias

from ..tools.backend_api import (
    BACKEND_PORT,
    LOCALHOST,
    Action,
    backend_request,
)
from ..tools.log import log_manager

LOGGING_COLORS = {
    logging.DEBUG: "grey66",
    logging.INFO: "green",
    logging.WARNING: "yellow",
    logging.ERROR: "red",
    logging.CRITICAL: "bold purple",
}

logging.getLogger().setLevel(logging.CRITICAL + 1)


class LogHandler(logging.Handler):
    def __init__(
        self, callback: Callable[[logging.LogRecord], None] | None = None
    ) -> None:
        super().__init__()
        self.callback = callback

    def emit(self, record: logging.LogRecord) -> None:
        if self.callback is not None:
            self.callback(record)


class BackendStatus:
    DEFAULT_REFRESH_RATE = 0.5
    DEFAULT_N_CONSOLE_LINES = 20  # Number of lines in the console view
    CONNECTIONS_MIN_HEIGHT = 2

    def __init__(self, input_args: list[str]) -> None:
        self.argument_parser = argparse.ArgumentParser()
        self.argument_parser.add_argument(
            "-l",
            "--live",
            type=float,
            default=False,
            help="Live mode, update status continuously at set refresh rate",
        )
        self.argument_parser.add_argument(
            "--log-lines",
            type=int,
            default=self.DEFAULT_N_CONSOLE_LINES,
            help="Number of log lines to display",
        )
        self.argument_parser.add_argument(
            "-a",
            "--address",
            type=str,
            default=LOCALHOST,
            help="Listening address, set it to the interface that will be used by the client",
        )
        self.argument_parser.add_argument(
            "-p", "--port", type=int, default=BACKEND_PORT
        )
        self.argument_parser.add_argument(
            "--log-level", type=str, choices=list(logging._nameToLevel.keys())
        )  # pyright: ignore[reportPrivateUsage]

        args = self.argument_parser.parse_args(input_args)

        self._n_console_lines = args.log_lines
        self.host: str = args.address
        self.port: int = args.port
        self.live_delay: float | None = args.live

        # Buffer for last N_CONSOLE_LINES log messages
        self.console_lines: list[str] = []
        self.console_lock = threading.Lock()  # Lock for thread safety
        log_manager.enable_backend_logging()
        self._start_time = time.time()
        self.conn: NamedConnection | None = None

        self._log_handler = LogHandler()

        self._new_console_line_r, self._new_console_line_w = Pipe()

    def run(self) -> None:
        self._log_handler.callback = self._add_line_status_screen
        logging.getLogger().addHandler(self._log_handler)
        adapter_table = None
        main_table = Table(box=None, padding=(0, 1))
        main_table.add_column(justify="right")
        main_table.add_column(justify="left")
        try:
            with Live(refresh_per_second=30) as live_display:
                while True:
                    if self.conn is None:
                        pid = ""
                        backend_status = "[red3]● Offline"
                        monitoring_connections = Table(
                            "",
                            box=None,
                            caption_justify="right",
                            show_header=False,
                            pad_edge=False,
                        )
                        adapter_table = Table("", box=None, caption_justify="right")

                    if self.conn is None:
                        try:
                            self.conn = NamedConnection(Client((self.host, self.port)))
                        except (ConnectionRefusedError, OSError):
                            self.conn = None
                            sleep(0.1)
                        else:
                            backend_request(self.conn.conn, Action.SET_ROLE_MONITORING)

                    if self.conn is not None:
                        backend_status = "[green3]● Online"

                        ready = wait([self.conn.conn, self._new_console_line_r])

                        if self._new_console_line_r in ready:
                            self._new_console_line_r.recv()
                        elif self.conn.conn in ready:
                            try:
                                event = self.conn.conn.recv()
                            except (ConnectionResetError, OSError, EOFError):
                                self.conn = None
                                continue
                            else:
                                action: Action = Action(event[0])
                                if action == Action.ENUMERATE_ADAPTER_CONNECTIONS:
                                    adapter_table = Table(
                                        "", box=None, caption_justify="right"
                                    )
                                    snapshot: dict[str, tuple[bool, list[str]]] = event[
                                        1
                                    ]
                                    unique_clients: set[str] = set()
                                    for _, (_, adapter_clients) in snapshot.items():
                                        unique_clients |= set(adapter_clients)
                                    for client in unique_clients:
                                        adapter_table.add_column(
                                            f"{client}", justify="center"
                                        )

                                    for adapter, (
                                        status,
                                        adapter_clients,
                                    ) in snapshot.items():
                                        status_indicator = (
                                            "[green3]●" if status else "[red3]●"
                                        )
                                        client_indicators: list[str] = []
                                        for client in unique_clients:
                                            client_indicators.append(
                                                "[green3]●[/]"
                                                if client in adapter_clients
                                                else "[red3]●[/]"
                                            )
                                        adapter_table.add_row(
                                            f"{adapter} {status_indicator}",
                                            *client_indicators,
                                        )
                                elif action == Action.ENUMERATE_MONITORING_CONNECTIONS:
                                    monitoring_connections = Table(
                                        "",
                                        box=None,
                                        caption_justify="right",
                                        show_header=False,
                                        pad_edge=False,
                                    )
                                    # Update monitoring connections
                                    monitoring_response: list[tuple[str, str]] = event[
                                        1
                                    ]

                                    for connection, desc in monitoring_response:
                                        if (
                                            connection
                                            == log_manager.backend_logger_conn_description()
                                            or connection == self.conn.local()
                                        ):
                                            style = ("grey50", "grey23")
                                        else:
                                            style = ("grey", "grey50")

                                        monitoring_connections.add_row(
                                            f"[{style[0]}]{connection}[/] [{style[1]}]({desc})[/]"
                                        )
                                    for _ in range(
                                        self.CONNECTIONS_MIN_HEIGHT
                                        - len(monitoring_response)
                                    ):
                                        monitoring_connections.add_row("")

                                elif action == Action.BACKEND_STATS:
                                    # Update Backend stats
                                    pid_response: int = cast(int, event[1])
                                    pid = str(pid_response)

                    main_table = Table(box=None, padding=(0, 1))
                    main_table.add_column(justify="right")
                    main_table.add_column(justify="left")
                    main_table.add_row("Status :", backend_status)
                    main_table.add_row("PID :", pid)
                    main_table.add_row("Logger connections :", monitoring_connections)
                    main_table.add_row(
                        "Adapter connections :",
                        (adapter_table if adapter_table is not None else Text("")),
                    )

                    # Build the console panel
                    with self.console_lock:
                        console_content = "\n".join(self.console_lines) or " "
                    console_panel = Panel(
                        console_content,
                        title="Console",
                        height=self._n_console_lines + 2,  # +2 for panel borders
                        border_style="dim",
                        padding=(0, 1),
                    )

                    # Group the main table and console panel vertically
                    group = Group(main_table, console_panel)
                    live_display.update(group)

                    if self.live_delay is not None:
                        sleep(self.live_delay)
                    else:
                        break
        except KeyboardInterrupt:
            pass

    def _format_record(self, record: logging.LogRecord) -> str:
        color = LOGGING_COLORS.get(record.levelno)
        relative_time = record.created - self._start_time

        line = f"[{color}]{relative_time:7.3f} {record.levelname:<8} {record.msg}[/]"
        if record.name == LoggerAlias.BACKEND.value:
            line = f"[bold]{line}[/bold]"
        return line

    def _add_line_status_screen(self, record: logging.LogRecord) -> None:
        """
        Add a log message to the console buffer, with color formatting by log level.
        Thread-safe.
        """
        formated_text = self._format_record(record)
        with self.console_lock:
            self.console_lines.append(formated_text)
            if len(self.console_lines) > self._n_console_lines:
                self.console_lines = self.console_lines[-self._n_console_lines :]
        self._new_console_line_w.send(b"\x00")



# backend_status.py
# 13.07.2025
# Sébastien Deriaz

import argparse
import logging
import time
from collections.abc import Callable
from multiprocessing.connection import Connection
from time import sleep
from typing import Any

from rich.console import Console
from rich.text import Text

from ..tools.backend_api import BACKEND_PORT, LOCALHOST
from ..tools.backend_logger import BackendLogger
from ..tools.log_settings import LoggerAlias

LOGGING_COLORS = {
    logging.DEBUG: "grey66",
    logging.INFO: "green",
    logging.WARNING: "yellow",
    logging.ERROR: "red",
    logging.CRITICAL: "bold purple",
}

logging.getLogger().setLevel(logging.CRITICAL + 1)


class LogHandler(logging.Handler):
    def __init__(
        self, callback: Callable[[logging.LogRecord], None] | None = None
    ) -> None:
        super().__init__()
        self.callback = callback

    def emit(self, record: logging.LogRecord) -> None:
        if self.callback is not None:
            self.callback(record)


class BackendConsole:
    def __init__(self, input_args: list[str]) -> None:
        self.argument_parser = argparse.ArgumentParser()
        self.argument_parser.add_argument(
            "-a",
            "--address",
            type=str,
            default=LOCALHOST,
            help="Listening address, set it to the interface that will be used by the client",
        )
        self.argument_parser.add_argument(
            "-p", "--port", type=int, default=BACKEND_PORT
        )
        self.argument_parser.add_argument(
            "-l", "--log-level", type=str, choices=list(logging._nameToLevel.keys())
        )  # pyright: ignore[reportPrivateUsage]

        args = self.argument_parser.parse_args(input_args)

        self.address = args.address
        self.port = args.port

        self._console = Console()

        self._log_handler = LogHandler()
        self._log_handler.callback = self._add_line

        logging.getLogger().addHandler(self._log_handler)
        self._start_time = time.time()
        self.conn: Connection[Any, Any] | None = None
        self._backend_logger = BackendLogger()

    def run(self) -> None:
        self._backend_logger.start()
        try:
            while True:
                sleep(1)
        except KeyboardInterrupt:
            pass

    def _format_record(self, record: logging.LogRecord) -> str:
        color = LOGGING_COLORS.get(record.levelno)
        relative_time = record.created - self._start_time

        line = f"[{color}]{relative_time:7.3f} {record.levelname:<8} {record.msg}[/]"
        if record.name == LoggerAlias.BACKEND.value:
            line = f"[bold]{line}[/bold]"
        elif record.name == LoggerAlias.BACKEND_LOGGER.value:
            line = f"[italic]{line}[/italic]"
        return line

    def _add_line(self, record: logging.LogRecord) -> None:
        formated_text = self._format_record(record)
        self._console.print(Text.from_markup(formated_text), markup=True)



# File : adapterbackend.py
# Author : Sébastien Deriaz
# License : GPL
#
# The adapter backend is the background class that manages communication
# with one particular device. It is always instanciated in a backend client

import logging
import socket
import time
from abc import ABC, abstractmethod
from collections.abc import Generator
from dataclasses import dataclass
from enum import Enum
from multiprocessing.connection import Connection
from threading import Thread
from typing import Protocol

from ...tools.backend_api import AdapterBackendStatus, Fragment
from ...tools.log_settings import LoggerAlias
from ..stop_condition import StopConditionType
from .descriptors import Descriptor
from .stop_condition_backend import (
    ContinuationBackend,
    StopConditionBackend,
    TotalBackend,
)


class HasFileno(Protocol):
    def fileno(self) -> int:
        return -1


Selectable = HasFileno | int

DEFAULT_STOP_CONDITION = None


class SocketReadException(Exception):
    pass


# STOP_DESIGNATORS = {
#     # "timeout": {
#     #     TimeoutType.RESPONSE: "TR",
#     #     TimeoutType.CONTINUATION: "TC",
#     #     TimeoutType.TOTAL: "TT",
#     # },
#     "stop_condition": {TerminationBackend: "ST", LengthBackend: "SL"},
#     "previous-buffer": "PB",
# }


class Origin(Enum):
    TIMEOUT = "timeout"
    STOP_CONDITION = "stop_condition"


class AdapterSignal:
    pass


class AdapterDisconnectedSignal(AdapterSignal):
    def __str__(self) -> str:
        return "Adapter disconnected"

    def __repr__(self) -> str:
        return self.__str__()


# @dataclass
# class AdapterReadInit(AdapterSignal):
#     end_delay : float | None

#     def __str__(self) -> str:
#         return f"Read init (end delay : {self.end_delay})"

#     def __repr__(self) -> str:
#         return self.__str__()


@dataclass
class AdapterResponseTimeout(AdapterSignal):
    identifier: int

    def __str__(self) -> str:
        return "Response timeout"

    def __repr__(self) -> str:
        return self.__str__()


@dataclass
class AdapterReadPayload(AdapterSignal):
    fragments: list[Fragment]
    stop_timestamp: float
    stop_condition_type: StopConditionType
    previous_read_buffer_used: bool
    response_timestamp: float | None
    # Only used by client and set by frontend
    response_delay: float | None = None

    def data(self) -> bytes:
        return b"".join([f.data for f in self.fragments])

    def __str__(self) -> str:
        return f"Read payload : {self.data()!r}"

    def __repr__(self) -> str:
        return self.__str__()


# This class holds a request made by the client to know if a response
# is received without the specified time frame
@dataclass
class ResponseRequest:
    timestamp: float
    identifier: int


def nmin(a: float | None, b: float | None) -> float | None:
    if a is None and b is None:
        return None
    elif a is None:
        return b
    elif b is None:
        return a
    else:
        return min(a, b)


class AdapterBackend(ABC):
    class ThreadCommands(Enum):
        STOP = b"0"

    class AdapterTimeoutEventOrigin(Enum):
        TIMEOUT = 0
        RESPONSE_REQUEST = 1

    def __init__(self, descriptor: Descriptor) -> None:
        """
        Adapter instance

        Parameters
        ----------
        timeout : float or Timeout instance
            Default timeout is Timeout(response=5, continuation=0.2, total=None)
        stop_condition : StopCondition or None
            Default to None
        """
        self._logger = logging.getLogger(LoggerAlias.ADAPTER_BACKEND.value)

        super().__init__()

        # TODO : Switch to multiple stop conditios
        self._stop_conditions: list[StopConditionBackend] = [
            ContinuationBackend(time=0.1)
        ]
        self.descriptor = descriptor
        self._thread: Thread | None = None
        self._status = AdapterBackendStatus.DISCONNECTED
        self._thread_commands_read, self._thread_commands_write = socket.socketpair()
        self.backend_signal: Connection | None = None
        self.fragments: list[Fragment] = []
        self._next_timeout_timestamp: float | None = None
        # _response_time indicates if the frontend asked for a read
        # None : No ask
        # float : Ask for a response to happen at the specified value at max
        self._response_request: ResponseRequest | None = None
        self._response_request_start: float | None = None

        self._first_fragment = True

        # Buffer for data that has been pulled from the queue but
        # not used because of termination or length stop condition
        self._previous_buffer = Fragment(b"", None)

        self._last_write_time = time.time()

    def set_stop_conditions(self, stop_conditions: list[StopConditionBackend]) -> None:
        """
        Overwrite the stop-condition

        Parameters
        ----------
        stop_condition : StopCondition
        """
        self._stop_conditions = stop_conditions

    def flush_read(self) -> bool:
        """
        Flush the input buffer
        """
        self._logger.debug("Flush")
        self._previous_buffer = Fragment(b"", None)
        self._response_request = None
        self.fragments = []
        for stop_condition in self._stop_conditions:
            stop_condition.flush_read()
        return True

    def previous_read_buffer_empty(self) -> bool:
        """
        Check whether the previous read buffer is empty

        Returns
        -------
        empty : bool
        """
        return self._previous_buffer.data == b""

    @abstractmethod
    def open(self) -> None:
        """
        Start communication with the device
        """
        pass

    def close(self) -> bool:
        """
        Stop communication with the device
        """
        self._status = AdapterBackendStatus.DISCONNECTED
        return True

    def write(self, data: bytes) -> bool:
        """
        Send data to the device

        Parameters
        ----------
        data : bytes or str
        """
        self._last_write_time = time.time()
        self._logger.debug(f"Write {repr(data)}")
        return True

    @abstractmethod
    def selectable(self) -> HasFileno | None:
        """
        Return an object with a fileno() method (e.g., socket, Connection) suitable for use with select/poll.
        """
        raise NotImplementedError

    @abstractmethod
    def _socket_read(self) -> Fragment:
        raise NotImplementedError

    def _fragments_to_string(self, fragments: list[Fragment]) -> str:
        if len(fragments) > 0:
            return "+".join(repr(f.data) for f in fragments)
        else:
            return str([])

    def on_socket_ready(self) -> Generator[AdapterSignal, None, None]:
        fragment = self._socket_read()
        if fragment.timestamp is not None and self._last_write_time is not None:
            fragment_delta_t = fragment.timestamp - self._last_write_time
        else:
            fragment_delta_t = float("nan")
        if fragment.data == b"":
            self.close()
            yield AdapterDisconnectedSignal()
        else:
            self._logger.debug(
                f"New fragment {fragment_delta_t:+.3f} {fragment}"
                + (" (first)" if self._first_fragment else "")
            )
            if self._status == AdapterBackendStatus.CONNECTED:
                t = time.time()

                # If there's a response request, disable it if there's a timeout stop condition
                # The stop-condition will do the job

                if self._response_request is not None:
                    for stop_condition in self._stop_conditions:
                        if isinstance(
                            stop_condition, (ContinuationBackend, TotalBackend)
                        ):
                            self._response_request = None
                            break

                while True:
                    if self._first_fragment:
                        self._first_fragment = False
                        self._read_start_timestamp = t
                        for stop_condition in self._stop_conditions:
                            stop_condition.initiate_read()

                    stop = False
                    kept = fragment

                    # Run each stop condition one after the other, if a stop is reached, stop evaluating
                    stop_condition_type: StopConditionType
                    for stop_condition in self._stop_conditions:
                        (
                            stop,
                            kept,
                            self._previous_buffer,
                            self._next_timeout_timestamp,
                        ) = stop_condition.evaluate(kept)
                        if stop:
                            stop_condition_type = stop_condition.type()
                            break

                    # if kept.data != b'':
                    #     self.fragments.append(kept)

                    self.fragments.append(kept)

                    if stop:
                        self._first_fragment = True
                        self._logger.debug(
                            f"Payload {self._fragments_to_string(self.fragments)} ({stop_condition_type.value})"
                        )
                        if (
                            self._response_request_start is None
                            or len(self.fragments) == 0
                        ):
                            response_delay = None
                        else:
                            if self.fragments[0].timestamp is None:
                                response_delay = None
                            else:
                                response_delay = (
                                    self.fragments[0].timestamp
                                    - self._response_request_start
                                )
                            self._response_request_start = None
                        yield AdapterReadPayload(
                            fragments=self.fragments,
                            stop_timestamp=t,
                            stop_condition_type=stop_condition_type,
                            previous_read_buffer_used=False,
                            response_timestamp=self.fragments[0].timestamp,
                            response_delay=response_delay,
                        )
                        self._next_timeout_timestamp = None  # Experiment !
                        self.fragments.clear()

                    if len(self._previous_buffer.data) > 0 and stop:
                        # If there's a previous buffer, put it in the fragment and loop again
                        # Only loop if there's a stop (oterwise a stop would never happen again)
                        fragment = self._previous_buffer

                    else:
                        # If not, quit now
                        break

        return None

    def start_read(self, response_time: float, identifier: int) -> None:
        """
        Start a read operation. This is a signal from the frontend. The only goal is to set the response time
        and tell the frontend if nothing arrives within a set time
        """
        self._response_request_start = time.time()
        self._logger.debug(f"Setup read [{identifier}] in {response_time:.3f} s")
        self._response_request = ResponseRequest(
            self._response_request_start + response_time, identifier
        )

    @abstractmethod
    def is_opened(self) -> bool:
        """
        Return True if adapter is opened, False otherwise
        """

    def __str__(self) -> str:
        return self.descriptor.__str__()

    def __repr__(self) -> str:
        return self.__str__()

    def on_timeout_event(self) -> AdapterSignal | None:
        t = time.time()

        if self._next_timeout_origin == self.AdapterTimeoutEventOrigin.TIMEOUT:
            self._next_timeout_timestamp = None
            self._first_fragment = True
            if self._response_request_start is None or len(self.fragments) == 0:
                response_delay = None
            else:
                if self.fragments[0].timestamp is not None:
                    response_delay = (
                        self.fragments[0].timestamp - self._response_request_start
                    )
                else:
                    response_delay = None
                self._response_request_start = None

            output = AdapterReadPayload(
                stop_timestamp=t,
                stop_condition_type=StopConditionType.TIMEOUT,
                previous_read_buffer_used=False,
                fragments=self.fragments,
                response_timestamp=(
                    self.fragments[0].timestamp if len(self.fragments) > 0 else None
                ),
                response_delay=response_delay,
            )
            # Reset response request
            if self._response_request is not None:
                self._response_request = None
                self._response_request_start = None
            # Clear all of the fragments
            self.fragments = []
            return output

        elif (
            self._next_timeout_origin == self.AdapterTimeoutEventOrigin.RESPONSE_REQUEST
        ):
            if self._response_request is not None:
                signal = AdapterResponseTimeout(self._response_request.identifier)
                self._response_request = None
                return signal

        return None

    def get_next_timeout(self) -> float | None:
        min_timestamp = None
        self._next_timeout_origin = None

        if self._next_timeout_timestamp is not None:
            min_timestamp = self._next_timeout_timestamp
            self._next_timeout_origin = self.AdapterTimeoutEventOrigin.TIMEOUT

        if self._response_request is not None:
            if (
                min_timestamp is None
                or self._response_request.timestamp < min_timestamp
            ):
                min_timestamp = self._response_request.timestamp
                self._next_timeout_origin = (
                    self.AdapterTimeoutEventOrigin.RESPONSE_REQUEST
                )
        return min_timestamp



# File : timed_queue.py
# Author : Sébastien Deriaz
# License : GPL
#
# The timed queue is a special queue that stores time information when elements are stored inside
# This is used to manage timing when reading data from devices

# import queue
# from time import time


# class TimedQueue:
#     def __init__(self) -> None:
#         self._queue = queue.Queue()

#     def put(self, fragment: bytes) -> None:
#         self._queue.put((time(), fragment))

#     def get(self, timeout: float | None) -> tuple[float, bytes]:
#         """
#         Return an element of the timed queue. Waits at most the amount of time specified by timeout

#         Parameters
#         ----------
#         timeout : float, None
#         """
#         try:
#             return self._queue.get(block=True, timeout=timeout)
#         except queue.Empty:  # No item after timeout
#             return None, None

#     def is_empty(self):
#         return self._queue.empty()

#     def clear(self):
#         with self._queue.mutex:
#             self._queue.queue.clear()
#             self._queue.all_tasks_done.notify_all()
#             self._queue.unfinished_tasks = 0



import argparse
from contextlib import nullcontext

from rich.console import Console

from syndesi.adapters.backend.backend import Backend
from syndesi.tools.backend_api import BACKEND_PORT, LOCALHOST


class BackendWrapper:
    def __init__(self, remaining_args: list[str]) -> None:
        argument_parser = argparse.ArgumentParser()

        argument_parser.add_argument(
            "-a",
            "--address",
            type=str,
            default=LOCALHOST,
            help="Listening address, set it to the interface that will be used by the client",
        )
        argument_parser.add_argument("-p", "--port", type=int, default=BACKEND_PORT)
        argument_parser.add_argument(
            "-s",
            "--shutdown-delay",
            type=int,
            default=None,
            help="Delay before the backend shutdowns automatically",
        )
        argument_parser.add_argument(
            "-q", "--quiet", default=False, action="store_true"
        )
        argument_parser.add_argument(
            "-v", "--verbose", default=False, action="store_true"
        )

        args = argument_parser.parse_args(remaining_args)

        self.address = args.address
        self.port = args.port
        self.shutdown_delay = args.shutdown_delay
        self.quiet = args.quiet
        self.verbose = args.verbose

    def run(self) -> None:
        console = Console()
        with (
            console.status(
                f"[bold green]Syndesi backend running on {self.address}", spinner="dots"
            )
            if not self.quiet
            else nullcontext()
        ):
            backend = Backend(
                host=self.address,
                port=self.port,
                backend_shutdown_delay=self.shutdown_delay,
            )
            try:
                backend.start()
            except KeyboardInterrupt:
                pass



# File : backend_api.py
# Author : Sébastien Deriaz
# License : GPL

from dataclasses import dataclass
from enum import Enum
from multiprocessing.connection import Connection, wait
from typing import Any, Protocol

from syndesi.tools.errors import BackendCommunicationError, BackendError

BACKEND_PORT = 59677
LOCALHOST = "127.0.0.1"

default_host = LOCALHOST
# AUTHKEY = b'syndesi'

# Backend protocol format
# (action, arguments)
# If the command succeeds, it is returned as :
# (action, other arguments)
# If the command fails, it is returned as :
# (error, error_description)

# The backend links the client with an adapter when SELECT_ADAPTER is sent along with an adapter descriptor

EXTRA_BUFFER_RESPONSE_TIME = 1

# Delay to let the adapter connect
DEFAULT_ADAPTER_OPEN_TIMEOUT = 0.5


class Action(Enum):
    # All adapters
    SELECT_ADAPTER = "select"
    OPEN = "open"  # (descriptor,stop_condition) -> ()
    CLOSE = "close"  # (descriptor,force) -> ()
    # FORCE_CLOSE = "force_close"  # (descriptor,) -> ()
    WRITE = "write"  # (descriptor,data) -> ()
    READ = "read"  # (descriptor,full_output,temporary_timeout,temporary_stop_condition) -> (data,metrics)
    SET_STOP_CONDITIONs = "set_stop_condition"  # (descriptor,stop_condition)
    FLUSHREAD = "flushread"
    START_READ = "start_read"  # Start a read (descriptor,response_time)
    RESPONSE_TIMEOUT = "response_timeout"

    # Signal
    ADAPTER_SIGNAL = "adapter_signal"

    # Other
    SET_ROLE_ADAPTER = (
        "set_role_adapter"  # Define the client as an adapter (exchange of data)
    )
    SET_ROLE_MONITORING = "set_role_monitoring"  # The client queries for backend info
    SET_ROLE_LOGGER = "set_role_logger"  # The client receives logs
    SET_LOG_LEVEL = "set_log_level"
    PING = "ping"
    STOP = "stop"

    # Backend debugger
    ENUMERATE_ADAPTER_CONNECTIONS = "enumerate_adapter_connections"
    ENUMERATE_MONITORING_CONNECTIONS = "enumerate_monitoring_connections"
    BACKEND_STATS = "backend_stats"

    # Errors
    ERROR_GENERIC = "error_generic"
    ERROR_UNKNOWN_ACTION = "error_unknown_action"
    ERROR_INVALID_REQUEST = "error_invalid_request"
    ERROR_ADAPTER_NOT_OPENED = "error_adapter_not_opened"
    ERROR_INVALID_ROLE = "error_invalid_role"
    ERROR_ADAPTER_DISCONNECTED = "error_adapter_disconnected"
    ERROR_BACKEND_DISCONNECTED = "error_backend_disconnected"
    ERROR_FAILED_TO_OPEN = "error_failed_to_open"


def is_action_error(action: Action) -> bool:
    return action.value.startswith("error_")


class BackendException(Exception):
    pass


class ValidFragment(Protocol):
    data: bytes


@dataclass
class Fragment:
    data: bytes
    timestamp: float | None

    def __str__(self) -> str:
        return f"{self.data!r}"

    def __repr__(self) -> str:
        return self.__str__()

    def __getitem__(self, key: slice) -> "Fragment":
        # if self.data is None:
        #     raise IndexError('Cannot index invalid fragment')
        return Fragment(self.data[key], self.timestamp)


BackendResponse = tuple[object, ...]


def frontend_send(conn: Connection, action: Action, *args: Any) -> bool:
    try:
        conn.send((action.value, *args))
    except (BrokenPipeError, OSError):
        return False
    else:
        return True


def backend_request(
    conn: Connection,
    action: Action,
    *args: Any,
    timeout: float = EXTRA_BUFFER_RESPONSE_TIME,
) -> BackendResponse:
    try:
        conn.send((action.value, *args))
    except (BrokenPipeError, OSError) as err:
        raise BackendCommunicationError("Failed to communicate with backend") from err
    else:
        ready = wait([conn], timeout=timeout)
        if conn not in ready:
            raise BackendCommunicationError(
                "Failed to receive backend response in time"
            )

        try:
            raw_response: object = conn.recv()
        except (EOFError, ConnectionResetError) as err:
            raise BackendCommunicationError(
                f"Failed to receive backend response to {action.value}"
            ) from err

        # Check if the response is correctly formatted
        if not (isinstance(raw_response, tuple) and isinstance(raw_response[0], str)):
            raise BackendCommunicationError(
                f"Invalid response received from backend : {raw_response}"
            )

        response_action: Action = Action(raw_response[0])
        arguments: tuple[Any, ...] = raw_response[1:]

        if is_action_error(response_action):
            if len(arguments) > 0:
                if isinstance(arguments[0], str):
                    error_message: str = arguments[0]
                else:
                    error_message = "failed to read error message"
            else:
                error_message = "Missing error message"
            raise BackendError(f"{response_action} : {error_message}")
        return arguments


backend_send = frontend_send


def raise_if_error(response: BackendResponse) -> None:
    action = Action(response[0])
    if is_action_error(action):
        if len(response) > 1:
            description = response[1]
        else:
            description = f"{action}"
        raise BackendException(f"{action.name}/{description}")
    return


class AdapterBackendStatus(Enum):
    DISCONNECTED = 0
    CONNECTED = 1


class ClientStatus(Enum):
    DISCONNECTED = 0
    CONNECTED = 1
